"""Unit test of `_collections` module."""

import collections
import logging
import unittest
from typing import Any, MutableMapping

import pytest

from yaslha._collections import OrderedCaseInsensitiveDict as oci_dict
from yaslha._collections import OrderedTupleOrderInsensitiveDict as toi_dict

logger = logging.getLogger("test_info")


class TestOrderedCaseInsensitiveDict(unittest.TestCase):
    """Unit test of `OrderedCaseInsensitiveDict`."""

    def setUp(self):
        self.d = oci_dict()  # type: oci_dict[Any, Any]
        self.d["FIRST"] = 100
        self.d[2] = None
        self.d[("Third", 0)] = (1, 2, 3)  # inner elements are not normalized
        self.d[1] = "AnB"

    def test_init_by_list(self):
        d = oci_dict(
            [("a", 10), ("B", 30), (None, "AnB"), (1, "300aBC")]
        )  # type: oci_dict[Any, Any]
        assert d["A"] == 10
        assert d["b"] == 30
        assert d["B"] == 30
        assert d[None] == "AnB"
        assert d[1] == "300aBC"
        assert list(d.keys()) == ["A", "B", None, 1]

    def test_init_by_keywords(self):
        d = oci_dict(
            FiRST=3, SeCoND=None, third=(1, 2, 3), lAST="AnB"
        )  # type: oci_dict[Any, Any]
        assert d["FIRST"] == 3
        assert d["second"] is None
        assert d["third"] == (1, 2, 3)
        assert d["last"] == "AnB"

    def test_contains(self):
        assert "first" in self.d
        assert "FIRst" in self.d
        assert "_first" not in self.d

        assert None not in self.d
        assert 2 in self.d
        assert 3 not in self.d

        assert ("Third", 0) in self.d
        assert ("third", 0) not in self.d

    def test_delitem(self):
        del self.d["FIrST"]
        del self.d[1]
        assert list(self.d.keys()) == [2, ("Third", 0)]

    def test_dict(self):
        d = dict(self.d)  # type: MutableMapping[Any, Any]
        assert len(d.keys()) == 4
        assert d["FIRST"] == self.d["first"]
        assert d[2] == self.d[2]
        assert d[("Third", 0)] == self.d[("Third", 0)]
        assert d[1] == self.d[1]

    def test_eq_self(self):
        other = oci_dict(
            [("fIRst", 100), (2, None), (("Third", 0), (1, 2, 3)), (1, "AnB")]
        )  # type: oci_dict[Any, Any]
        assert self.d == other

        del other["first"]
        other["first"] = 100  # reorder

        assert self.d != other
        assert self.d == dict(other)
        assert dict(self.d) == other
        assert dict(self.d) == dict(other)

    def test_eq_ordered_dict(self):
        other = collections.OrderedDict(
            [("FIRST", 100), (2, None), (("Third", 0), (1, 2, 3)), (1, "AnB")]
        )
        assert self.d == other

        del other["FIRST"]
        other["FIRST"] = 100  # reorder

        assert self.d != other
        assert self.d == dict(other)
        assert dict(self.d) == other
        assert dict(self.d) == dict(other)

    def test_getitem_and_get(self):
        assert self.d["FiRST"] == 100
        assert self.d[2] is None
        assert self.d[("Third", 0)] == (1, 2, 3)
        with pytest.raises(KeyError):
            self.d[("third", 0)]  # inner elements are case sensitive

        assert self.d.get("FiRSt") == 100
        assert self.d.get(2) is None
        assert self.d.get(("Third", 0)) == (1, 2, 3)
        assert self.d.get(("third", 0), "NotFound") == "NotFound"
        assert self.d.get(123, 456) == 456

    def test_len(self):
        assert len(self.d) == 4

    def test_clear(self):
        self.d.clear()
        assert len(self.d) == 0

    def test_move_to_end(self):
        self.d.move_to_end("FIRST", last=True)  # case insensitive
        assert list(self.d.keys()) == [2, ("Third", 0), 1, "FIRST"]

        self.d.move_to_end(("Third", 0), last=False)
        assert list(self.d.keys()) == [("Third", 0), 2, 1, "FIRST"]

    def test_fromkeys(self):
        d = oci_dict.fromkeys(["a", "B", "thiRD", ("M", "n"), 3], True)
        assert list(d.keys()) == ["A", "B", "THIRD", ("M", "n"), 3]
        assert d["A"]
        assert d["B"]
        assert d["Third"]
        assert d[("M", "n")]
        assert d[3]

    def test_copy(self):
        d = self.d.copy()
        assert d == self.d
        d.move_to_end("first")
        assert d != self.d

    def test_keys_and_reversed(self):
        expected = ["FIRST", 2, ("Third", 0), 1]
        assert list(self.d.keys()) == expected
        assert list(reversed(self.d)) == list(reversed(expected))

    def test_values(self):
        assert list(self.d.values()) == [100, None, (1, 2, 3), "AnB"]

    def test_items(self):
        ks, vs = [], []
        for k, v in self.d.items():
            ks.append(k)
            vs.append(v)
        assert ks == list(self.d.keys())
        assert vs == list(self.d.values())

    def test_update(self):
        other = {"FIRST": 9, "seCond": 2, 2: True}
        self.d.update(other)
        assert self.d["fiRSt"] == 9
        assert self.d["Second"] == 2
        assert self.d[2] is True
        assert self.d[1] == "AnB"

    def test_pop(self):
        assert self.d.pop("fiRSt") == 100
        assert self.d.pop(2) is None
        with pytest.raises(KeyError):
            self.d.pop("first")
        with pytest.raises(KeyError):
            self.d.pop(100)
        assert self.d.pop(("Third", 0)) == (1, 2, 3)
        assert len(self.d) == 1

    def test_pop_with_default(self):
        assert self.d.pop("fiRSt", 9) == 100
        assert self.d.pop(2, default=9) is None
        assert self.d.pop("first", 12345) == 12345
        assert self.d.pop(100, default=12345) == 12345
        assert len(self.d) == 2

    def test_popitem(self):
        assert self.d.popitem(last=True) == (1, "AnB")
        assert self.d.popitem(last=False) == ("FIRST", 100)
        assert len(self.d) == 2


class TestOrderedTupleOrderInsensitiveDict(unittest.TestCase):
    """Unit test of `OrderedTupleOrderInsensitiveDict`."""

    def setUp(self):
        self.d = toi_dict()  # type: toi_dict[Any, Any]
        self.d[1, 1, 1, 0] = 100
        self.d[300] = None
        self.d[(2, 1), (4, 2), (2, 3)] = (1, 2, 3)  # inner elements are not ordered
        self.d[4, 3, 2, 1] = "AnB"

    def test_init_by_list(self):
        d = toi_dict(
            [
                ((0, 1, 1, 1), 100),
                (300, None),
                (((2, 3), (2, 1), (4, 2)), (1, 2, 3)),
                ((1, 3, 2, 4), "AnB"),
            ]
        )  # type: toi_dict[Any, Any]
        assert self.d == d
        assert d[1, 0, 1, 1] == 100
        assert d[0, 1, 1, 1] == 100
        assert d[300] is None
        assert d[(2, 1), (2, 3), (4, 2)] == (1, 2, 3)
        assert d[(4, 2), (2, 1), (2, 3)] == (1, 2, 3)
        assert d[1, 2, 3, 4] == "AnB"
        assert d[4, 2, 1, 3] == "AnB"
        assert list(d.keys()) == [
            (0, 1, 1, 1),
            300,
            ((2, 1), (2, 3), (4, 2)),
            (1, 2, 3, 4),
        ]

        with pytest.raises(KeyError):
            d[(2, 4), (1, 2), (2, 3)]

    def test_init_by_keywords(self):
        d = toi_dict(first=3, second=None)  # type: toi_dict[Any, Any]
        assert d["first"] == 3
        assert d["second"] is None

    def test_contains(self):
        assert (0, 1, 1, 1) in self.d
        assert (1, 0, 1, 1) in self.d
        assert (0, 0, 0, 0) not in self.d
        assert (1, 1, 1, 1) not in self.d

        assert None not in self.d
        assert 10 not in self.d

        assert ((2, 1), (2, 3), (4, 2)) in self.d
        assert ((1, 2), (2, 3), (4, 2)) not in self.d
        assert ((1, 2), (2, 3), (2, 4)) not in self.d

        # a bit tricky...
        assert 300 in self.d
        assert (300,) not in self.d
        assert (1, 2, 3, 4) in self.d

    def test_delitem(self):
        del self.d[0, 1, 1, 1]
        del self.d[300]
        del self.d[(4, 2), (2, 1), (2, 3)]
        assert list(self.d.keys()) == [(1, 2, 3, 4)]

    def test_dict(self):
        d = dict(self.d)  # type: MutableMapping[Any, Any]
        assert len(d.keys()) == 4
        for k in d.keys():
            assert d[k] == self.d[k]

    def test_eq_self(self):
        other = toi_dict(
            [
                ((0, 1, 1, 1), 100),
                (300, None),
                (((2, 3), (2, 1), (4, 2)), (1, 2, 3)),
                ((1, 2, 3, 4), "AnB"),
            ]
        )  # type: toi_dict[Any, Any]
        assert self.d == other

        del other[1, 1, 0, 1]
        other[1, 0, 1, 1] = 100  # reorder

        assert self.d != other
        assert self.d == dict(other)
        assert dict(self.d) == other
        assert dict(self.d) == dict(other)

    def test_eq_ordered_dict(self):
        other = collections.OrderedDict(
            [
                ((0, 1, 1, 1), 100),
                (300, None),
                (((2, 1), (2, 3), (4, 2)), (1, 2, 3)),
                ((1, 2, 3, 4), "AnB"),
            ]
        )
        assert self.d == other

        del other[0, 1, 1, 1]
        other[0, 1, 1, 1] = 100  # reorder

        assert self.d != other
        assert self.d == dict(other)
        assert dict(self.d) == other
        assert dict(self.d) == dict(other)

    def test_getitem_and_get(self):
        assert self.d[1, 1, 1, 0] == 100
        assert self.d[0, 1, 1, 1] == 100
        assert self.d[300] is None
        assert self.d[(2, 3), (4, 2), (2, 1)] == (1, 2, 3)
        with pytest.raises(KeyError):
            self.d[(1, 2), (2, 3), (2, 4)]  # inner elements are order preserved

        assert self.d.get((1, 1, 0, 1)) == 100
        assert self.d.get(300) is None
        assert self.d.get((4, 3, 2, 1)) == "AnB"
        assert self.d.get((4, 3, 2, 1), default=True) == "AnB"
        assert self.d.get((4, 3, 2, 1), True) == "AnB"
        assert self.d.get((1, 2, 3, 5)) is None
        assert self.d.get((1, 2, 3, 5), default=True) is True
        assert self.d.get((1, 2, 3, 5), True) is True

        assert self.d[0, 1, 1, 1] == 100
        assert self.d[0, 1, 1, 1] == 100

    def test_len(self):
        assert len(self.d) == 4

    def test_clear(self):
        self.d.clear()
        assert len(self.d) == 0

    def test_move_to_end(self):
        self.d.move_to_end((1, 1, 1, 0), last=True)  # case insensitive
        assert list(self.d.keys()) == [
            300,
            ((2, 1), (2, 3), (4, 2)),
            (1, 2, 3, 4),
            (0, 1, 1, 1),
        ]

        self.d.move_to_end(((2, 3), (4, 2), (2, 1)), last=False)
        assert list(self.d.keys()) == [
            ((2, 1), (2, 3), (4, 2)),
            300,
            (1, 2, 3, 4),
            (0, 1, 1, 1),
        ]

    def test_fromkeys(self):
        d = toi_dict.fromkeys([(1, 0), 3, (-1, 1, 0), (3, 2, -3)], True)
        assert list(d.keys()) == [(0, 1), 3, (-1, 0, 1), (-3, 2, 3)]
        assert d[1, 0]
        assert d[3]
        assert d[0, 1, -1]
        assert d[2, 3, -3]

    def test_copy(self):
        d = self.d.copy()
        assert d == self.d
        d.move_to_end((1, 0, 1, 1))
        assert d != self.d

    def test_keys_and_reversed(self):
        expected = [(0, 1, 1, 1), 300, ((2, 1), (2, 3), (4, 2)), (1, 2, 3, 4)]
        assert list(self.d.keys()) == expected
        assert list(reversed(self.d)) == list(reversed(expected))

    def test_values(self):
        assert list(self.d.values()) == [100, None, (1, 2, 3), "AnB"]

    def test_items(self):
        ks, vs = [], []
        for k, v in self.d.items():
            ks.append(k)
            vs.append(v)
        assert ks == list(self.d.keys())
        assert vs == list(self.d.values())

    def test_update(self):
        other = {(1, 0, 1, 1): 9, 300: 2, (3, 2, 1): True}
        self.d.update(other)
        assert self.d[1, 1, 0, 1] == 9
        assert self.d[300] == 2
        assert self.d[2, 3, 1] is True
        assert self.d[(2, 3), (2, 1), (4, 2)] == (1, 2, 3)
        assert self.d[1, 4, 2, 3] == "AnB"
        assert len(self.d) == 5

    def test_pop(self):
        assert self.d.pop((1, 0, 1, 1)) == 100
        assert self.d.pop(300) is None
        with pytest.raises(KeyError):
            self.d.pop((1, 1, 1, 0))
        with pytest.raises(KeyError):
            self.d.pop(300)
        assert self.d.pop((1, 2, 4, 3)) == "AnB"
        assert len(self.d) == 1

    def test_pop_with_default(self):
        assert self.d.pop((1, 0, 1, 1), 9) == 100
        assert self.d.pop(300, default=9) is None
        assert self.d.pop((0, 1, 1, 1), 12345) == 12345
        assert self.d.pop((1, 0, 1, 1), default=12345) == 12345
        assert len(self.d) == 2

    def test_popitem(self):
        assert self.d.popitem(last=True) == ((1, 2, 3, 4), "AnB")
        assert self.d.popitem(last=False) == ((0, 1, 1, 1), 100)
        assert len(self.d) == 2



"""Parsers for SLHA package.

Parsers for SLHA-format, JSON-format, and YAML-format are provided.
"""

import logging
from typing import Any, List, Optional, Type, Union

import yaslha.line
import yaslha.slha
from yaslha.block import AbsBlock, Block, Decay, InfoBlock

SLHAParserStatesType = Union[None, Block, InfoBlock, Decay]


logger = logging.getLogger(__name__)


class SLHAParser:
    """SLHA-format file parser."""

    def __init__(self, **kw: Any) -> None:
        self.processing = None  # type: SLHAParserStatesType

    def _parse_line(self, line: str) -> Optional[yaslha.line.AbsLine]:
        if not line.strip():
            return None  # empty line will be ignored
        if isinstance(self.processing, InfoBlock):
            classes = [
                yaslha.line.BlockHeadLine,
                yaslha.line.DecayHeadLine,
                yaslha.line.InfoLine,
                yaslha.line.CommentLine,
            ]  # type: List[Type[yaslha.line.AbsLine]]
        elif isinstance(self.processing, Block):
            classes = [
                yaslha.line.BlockHeadLine,
                yaslha.line.DecayHeadLine,
                yaslha.line.NoIndexLine,
                yaslha.line.OneIndexLine,
                yaslha.line.TwoIndexLine,
                yaslha.line.ThreeIndexLine,
                yaslha.line.DecayLine,  # for extensions
                yaslha.line.CommentLine,
            ]
        elif isinstance(self.processing, Decay):
            classes = [
                yaslha.line.BlockHeadLine,
                yaslha.line.DecayHeadLine,
                yaslha.line.DecayLine,
                yaslha.line.CommentLine,
            ]
        elif self.processing is None:
            classes = [
                yaslha.line.BlockHeadLine,
                yaslha.line.DecayHeadLine,
                yaslha.line.CommentLine,
            ]
        else:
            logger.critical("Unexpected state: %s", self.processing)
            raise RuntimeError

        for c in classes:
            obj = c.construct(line)
            if obj:
                return obj
        raise ValueError(line)

    def parse(self, text: str) -> yaslha.slha.SLHA:
        """Parse SLHA format text and return SLHA object."""
        self.processing = None
        slha = yaslha.slha.SLHA()
        comment_lines = []  # type: List[str]

        for line in text.splitlines():
            try:
                obj = self._parse_line(line)
                if obj is None:
                    continue
            except ValueError:
                logger.warning("Unrecognized line: %s", line)
                continue

            # comment handling
            if isinstance(obj, yaslha.line.CommentLine):
                comment_lines.append(obj.comment)
                continue
            elif isinstance(obj, yaslha.line.AbsLine):
                obj.pre_comment = comment_lines
                comment_lines = []
            else:
                raise NotImplementedError(obj)

            # line handling
            if isinstance(obj, yaslha.line.BlockHeadLine):
                self.processing = AbsBlock.new(obj)
                assert self.processing is not None
                slha.add_block(self.processing)
            elif isinstance(obj, yaslha.line.DecayHeadLine):
                self.processing = Decay(obj)
                assert self.processing is not None
                slha.add_block(self.processing)
            elif isinstance(obj, yaslha.line.InfoLine):
                if not isinstance(self.processing, InfoBlock):
                    logger.critical("InfoLine found outside of INFO block: %s", line)
                    raise ValueError(self.processing)
                self.processing.append_line(obj)
            elif isinstance(obj, yaslha.line.ValueLine):
                if self.processing is None:
                    logger.critical("ValueLine found outside of block: %s", line)
                    raise ValueError(self.processing)
                self.processing.update_line(obj)
            else:
                raise TypeError(obj)

        # tail comments
        slha.tail_comment = comment_lines
        self.processing = None
        return slha



"""Scripts of this package."""

import enum
import logging
import re
import sys
from typing import (  # noqa: F401
    Any,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Type,
    Union,
)

import click
import coloredlogs

import yaslha
import yaslha.dumper

logger = logging.getLogger(__name__)


ACCEPTED_TYPES = ["SLHA", "YAML", "JSON"]  # decided to use capital letters


class _Choice(click.Choice):
    def __init__(self, choices: Union[Sequence[str], Type[enum.Enum]]) -> None:
        if isinstance(choices, enum.EnumMeta):
            self.keys = {c.name.upper(): c for c in choices}
        else:
            self.keys = {c.upper(): c.upper() for c in choices}  # type: ignore
        super().__init__(self.keys.keys())  # type: ignore

    def convert(self, value, param, ctx):  # type: ignore
        return self.keys[super().convert(value.upper(), param, ctx)]


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.version_option(
    yaslha.__version__, "-V", "--version", prog_name=yaslha.__pkgname__
)
def main() -> None:
    """Handle SLHA format data."""
    coloredlogs.install(logger=logging.getLogger(), fmt="%(levelname)8s %(message)s")


@main.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.option(
    "--input-type",
    type=_Choice(["AUTO", *ACCEPTED_TYPES]),
    default="AUTO",
    show_default=True,
    help="(JSON/YAML input is not yet implemented.)",
    hidden=True,  # until other parsers are implemented
)
@click.option(
    "--output-type",
    type=_Choice(ACCEPTED_TYPES),
    default="SLHA",
    show_default=True,
    help="Output format.",
)
@click.argument("input", type=click.Path(exists=True, dir_okay=False), required=False)
@click.argument("output", type=click.Path(dir_okay=False), required=False)
@click.option("-S", "input_type", flag_value="SLHA", hidden=True)
@click.option("-J", "input_type", flag_value="JSON", hidden=True)
@click.option("-Y", "input_type", flag_value="YAML", hidden=True)
@click.option("-s", "output_type", flag_value="SLHA", hidden=True)
@click.option("-j", "output_type", flag_value="JSON", hidden=True)
@click.option("-y", "output_type", flag_value="YAML", hidden=True)
@click.option(
    "--comments",
    type=_Choice(yaslha.dumper.CommentsPreserve),
    default="NONE",
    show_default=True,
    help="Comment types to keep.",
)
@click.option(
    "--blocks",
    type=_Choice(yaslha.dumper.BlocksOrder),
    default="DEFAULT",
    help="Order of blocks.",
)
@click.option(
    "--values",
    type=_Choice(yaslha.dumper.ValuesOrder),
    default="DEFAULT",
    help="Order of values.",
)
def convert(**kwargs):
    # type: (Any)->None
    """Convert or reformat SLHA data files.

    Currently this package converts SLHA input to SLHA, JSON, or YAML format.
    Text from the standard input (STDIN) is used if no file is specified.
    Shorthanded options -s -y -j are available for --output-type option.
    """
    input_type = kwargs["input_type"] or "AUTO"
    output_type = kwargs["output_type"] or "SLHA"

    if kwargs["input"]:
        with open(kwargs["input"]) as f:
            input_string = f.read()
    else:
        logger.warning("Reading from STDIN...")
        input_string = sys.stdin.read()
    slha = yaslha.parse(input_string, input_type=input_type)

    output_string = yaslha.dump(
        slha=slha,
        output_type=output_type,
        comments_preserve=yaslha.dumper.CommentsPreserve(kwargs["comments"]),
        blocks_order=yaslha.dumper.BlocksOrder(kwargs["blocks"]),
        values_order=yaslha.dumper.ValuesOrder(kwargs["values"]),
    )

    if kwargs["output"]:
        with open(kwargs["output"], "w") as f:
            f.write(output_string)
    else:
        print(output_string)


@main.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.option(
    "-e", is_flag=True, default=False, help="Read the last data to merge from STDIN." ""
)
@click.argument(
    "input", nargs=-1, type=click.Path(exists=True, dir_okay=False), required=True
)
@click.pass_context  # for help
def merge(ctx, **kwargs):
    # type: (click.core.Context, Any)->None
    """Merge SLHA files.

    Input and output are in SLHA format. Duplicated data are overwritten by
    later files. If -e option is specified, input from standard input (STDIN)
    is used as the last data.
    """
    slha = yaslha.slha.SLHA()
    for i in kwargs["input"]:
        with open(i) as f:
            slha.merge(yaslha.parse(f.read()))
    if kwargs["e"]:
        slha.merge(yaslha.parse(sys.stdin.read()))

    if not (slha.blocks or slha.decays):
        click.echo(ctx.get_usage())
        ctx.exit(1)

    output_string = yaslha.dump(
        slha=slha,
        output_type="SLHA",
        comments_preserve=yaslha.dumper.CommentsPreserve.ALL,
        blocks_order=yaslha.dumper.BlocksOrder.KEEP,
        values_order=yaslha.dumper.ValuesOrder.KEEP,
    )
    print(output_string)


@main.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.argument("blocks", type=str, required=True)
@click.argument("input", type=click.Path(exists=True, dir_okay=False), required=False)
@click.pass_context  # for help
def extract(ctx, **kwargs):
    # type: (click.core.Context, Any)->None
    """Extract specified blocks from SLHA file.

    BLOCKS is a comma-separated list of block names (or decay-block PIDs) to
    extract and INPUT is the input SLHA file. If INPUT is not specified, data
    is read from standard input (STDIN).
    """
    blocks = kwargs["blocks"].split(",")
    if not blocks:
        click.echo("No blocks specified.")
        exit(1)

    if kwargs["input"]:
        with open(kwargs["input"]) as f:
            input_string = f.read()
    else:
        input_string = sys.stdin.read()
    slha = yaslha.parse(input_string)
    dumper = yaslha.dumper.SLHADumper()

    output_list = []  # type: List[Sequence[str]]
    for block in blocks:
        if re.match(r"^\d+$", block):
            try:
                output_list.append(dumper.dump_block(slha.decays[int(block)]))
            except KeyError:
                click.echo("DECAY block for PID {} not found.".format(block))
                exit(1)
        else:
            try:
                output_list.append(dumper.dump_block(slha.blocks[block]))
            except KeyError:
                click.echo("Block {} not found".format(block.upper()))
                exit(1)
    output_string = "\n".join("\n".join(block) for block in output_list)
    print(output_string)



"""Package to handle SLHA-format files and data."""

import pathlib
from typing import Any, Optional, Union

import yaslha.block
import yaslha.comment
import yaslha.config
import yaslha.dumper
import yaslha.line
import yaslha.parser
import yaslha.slha

__pkgname__ = "yaslha"
__version__ = "0.3.2"
__author__ = "Sho Iwamoto / Misho"
__license__ = "MIT"

cfg = yaslha.config.Config()


def parse(text, input_type="AUTO", parser=None, **kwargs):
    # type: (str, str, Any, Any)->yaslha.slha.SLHA
    """Parse a text to return an SLHA object."""
    if parser is None:
        if input_type.upper() == "AUTO":
            # TODO: implement auto-parser
            parser = yaslha.parser.SLHAParser(**kwargs)
        elif input_type.upper() == "JSON":
            raise NotImplementedError
        elif input_type.upper() == "YAML":
            raise NotImplementedError
        else:
            parser = yaslha.parser.SLHAParser(**kwargs)
    return parser.parse(text)


def dump(slha, output_type="SLHA", dumper=None, **kwargs):
    # type: (yaslha.slha.SLHA, str, Optional[yaslha.dumper.AbsDumper], Any)->str
    """Output a dumped string of an SLHA object."""
    if dumper is None:
        if output_type.upper() == "JSON":
            dumper = yaslha.dumper.JSONDumper(**kwargs)
        elif output_type.upper() == "YAML":
            dumper = yaslha.dumper.YAMLDumper(**kwargs)
        else:
            dumper = yaslha.dumper.SLHADumper(**kwargs)
    return dumper.dump(slha)


def parse_file(path, **kwargs):
    # type: (Union[str, pathlib.Path], Any)->yaslha.slha.SLHA
    """Parse a file to return an SLHA object."""
    if isinstance(path, str):
        path = pathlib.Path(path)
    return parse(path.read_text(), **kwargs)


def dump_file(data, path, **kwargs):
    # type: (yaslha.slha.SLHA, Union[str, pathlib.Path], Any)->None
    """Write into a file a dumped string of an SLHA object."""
    with open(str(path), "w") as f:
        f.write(dump(data, **kwargs))



"""Module of a class to handle comment interface."""

from typing import Any, Generic, List, Optional, TypeVar, Union, overload

from typing_extensions import Literal

from yaslha._line import DecayKeyType, InfoKeyType, KeyType
from yaslha.block import GenericBlock

KTG = TypeVar("KTG", KeyType, InfoKeyType, DecayKeyType)
CT = TypeVar("CT", str, List[str])
HEAD = Literal["head"]

class CommentInterface(Generic[KTG, CT]):
    _block: "GenericBlock[KTG, CT]"
    _pre: "PreCommentInterface[KTG, CT]"
    def __init__(self, block: "GenericBlock[KTG, CT]") -> None: ...
    @property
    def pre(self) -> "PreCommentInterface[KTG, CT]": ...
    @overload
    def __getitem__(self, key: HEAD) -> str: ...
    @overload
    def __getitem__(self, key: KTG) -> CT: ...
    # def __getitem__(self, key: Union[KTG, HEAD]) -> Union[CT, str]: ...
    @overload
    def __setitem__(self, key: HEAD, value: Optional[str]) -> None: ...
    @overload
    def __setitem__(self, key: KTG, value: Optional[CT]) -> None: ...
    # def __setitem__(self, key: Union[KTG, HEAD], value: Union[CT, str]) -> None: ...
    @overload
    def __call__(self, key: HEAD, **kw: Any) -> str: ...
    @overload
    def __call__(self, key: KTG, **kw: Any) -> CT: ...
    # def __call__(self, key: Union[KTG, HEAD], **kw: Any) -> Union[CT, str]: ...

class PreCommentInterface(Generic[KTG, CT]):
    _block: "GenericBlock[KTG, CT]"
    def __init__(self, block: "GenericBlock[KTG, CT]"): ...
    def __getitem__(self, key: Union[KTG, HEAD]) -> List[str]: ...
    def __setitem__(
        self, key: Union[KTG, HEAD], value: Optional[List[str]]
    ) -> None: ...
    def __call__(self, key: Union[KTG, HEAD], **kw: Any) -> List[str]: ...



"""Configuration handlers."""

import collections.abc
import configparser
import enum
import os
import pathlib
from typing import Any, List, MutableMapping, Type, TypeVar

CONFIG_FILES = [
    str(pathlib.Path(__file__).with_name("yaslha.cfg.default")),
    os.path.expanduser("~/.yaslha.cfg"),
    "yaslha.cfg",
]  # latter overrides former

EnumType = TypeVar("EnumType", bound=enum.Enum)


class SectionWrapper:
    """A wrapper class of `configparser.SectionProxy`."""

    def __init__(self, data: configparser.SectionProxy) -> None:
        self._data = data  # type: configparser.SectionProxy
        self.override = {}  # type: MutableMapping[str, Any]

    def __getattr__(self, name: str) -> Any:
        return self._data.__getattribute__(name)

    def __getitem__(self, key: str) -> Any:
        if key in self.override:
            return self.override[key]
        if key in self._data:
            return self._data[key]
        raise KeyError(key)

    def get_enum(self, key: str, enum_class: Type[EnumType]) -> EnumType:
        """Get an item as an Enum-class object."""
        if key in self.override:
            return self.override[key]  # type: ignore
        if key in self._data:
            value = self._data[key].lower()
            for i in enum_class:
                if i.name.lower() == value:
                    return i
        raise KeyError(key)

    def get_list(self, key: str) -> List[str]:
        """Get a List[str] object."""
        key_for_a_list = "{}@list".format(key)
        if key in self.override:
            value = self.override[key]
        elif key_for_a_list in self._data:
            value = self._data[key_for_a_list]
        else:
            raise KeyError(key)
        if isinstance(value, str):
            return [v for v in value.split(" ") if v]
        elif isinstance(value, collections.abc.Sequence):
            return [str(v) for v in value]
        else:
            raise TypeError(value)


class Config(configparser.ConfigParser):
    """Dictionary to store the configurations."""

    def __init__(self) -> None:
        super().__init__(inline_comment_prefixes="#")
        super().read(CONFIG_FILES)

    def __getitem__(self, key: Any) -> Any:
        return SectionWrapper(super().__getitem__(key))



"""Dumpers to write SLHA data in various format."""

import enum
import json
import re
from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from typing import (
    Any,
    ClassVar,
    List,
    Mapping,
    MutableMapping,
    Sequence,
    TypeVar,
    Union,
)

import ruamel.yaml

import yaslha
import yaslha.config
import yaslha.line
import yaslha.utility
from yaslha._line import format_comment
from yaslha.block import Block, Decay, InfoBlock

BlockLike = Union[Block, InfoBlock, Decay]
T = TypeVar("T")


@enum.unique
class BlocksOrder(enum.Enum):
    """Options for block ordering."""

    DEFAULT = 0
    KEEP = 1
    ABC = 2


@enum.unique
class ValuesOrder(enum.Enum):
    """Options for value ordering."""

    DEFAULT = 0
    KEEP = 1
    SORTED = 2


@enum.unique
class CommentsPreserve(enum.Enum):
    """Options for comment handling."""

    NONE = 0
    TAIL = 1
    ALL = 2

    @property
    def keep_line(self):
        # type: ()->bool
        """Return if to keep line-level comments."""
        return self == CommentsPreserve.ALL

    @property
    def keep_tail(self):
        # type: ()->bool
        """Return if to keep tail comments of value lines."""
        return self != CommentsPreserve.NONE


class AbsDumper(metaclass=ABCMeta):
    """Abstract class for YASLHA dumpers."""

    @abstractmethod
    def _read_config(self, sw: yaslha.config.SectionWrapper) -> None:
        self._config = {
            "blocks_order": sw.get_enum("blocks_order", BlocksOrder),
            "values_order": sw.get_enum("values_order", ValuesOrder),
            "comments_preserve": sw.get_enum("comments_preserve", CommentsPreserve),
        }  # type: MutableMapping[str, Any]

    @abstractmethod
    def config(self, k: str) -> Any:
        """Get a current value of configuration."""
        return self._config[k]

    @abstractmethod
    def set_config(self, k: str, v: Any) -> None:
        """Set configuration."""
        if any(
            [
                k == "blocks_order" and not isinstance(v, BlocksOrder),
                k == "values_order" and not isinstance(v, ValuesOrder),
                k == "comments_preserve" and not isinstance(v, CommentsPreserve),
            ]
        ):
            raise TypeError(k, v)
        self._config[k] = v

    @abstractmethod
    def __init__(self, **kw: Any) -> None:
        pass

    @abstractmethod
    def dump(self, slha: "yaslha.slha.SLHA") -> str:
        """Return dumped string of an SLHA object."""

    def _blocks_sorted(self, slha):
        # type: (yaslha.slha.SLHA)->List[Union[Block, InfoBlock]]
        slha.normalize(decays=False)
        if self.config("blocks_order") == BlocksOrder.KEEP:
            return list(slha.blocks.values())
        block_names = list(slha.blocks.keys())
        if self.config("blocks_order") == BlocksOrder.ABC:
            block_names.sort()
        else:
            block_names = yaslha.utility.sort_blocks_default(block_names)
        return [slha.blocks[name] for name in block_names]

    def _decays_sorted(self, slha):
        # type: (yaslha.slha.SLHA)->List[Decay]
        slha.normalize(blocks=False)
        if self.config("values_order") == ValuesOrder.KEEP:
            return list(slha.decays.values())
        pids = list(slha.decays.keys())
        if self.config("values_order") == ValuesOrder.SORTED:
            pids.sort()
        else:
            pids = yaslha.utility.sort_pids_default(pids)
        return [slha.decays[pid] for pid in pids]

    def _block_lines_ordered(self, block):
        # type: (BlockLike)->Sequence[yaslha.line.AbsLine]
        sort = self.config("values_order") != ValuesOrder.KEEP
        if (
            isinstance(block, Block)
            and self.config("values_order") == ValuesOrder.DEFAULT
            and block.name == "MASS"
        ):
            keys = yaslha.utility.sort_pids_default(list(block.keys()))
            return [block._data[k] for k in keys]
        else:
            return [line for _, line in block._lines(sort=sort)]

    @staticmethod
    def _document_out(lines: Sequence[str]) -> List[str]:
        return [
            "#"
            if not lines
            else line.replace(" ", "#", 1)
            if line.startswith(" ")
            else "#" + line.replace("  ", " ", 1)
            for line in lines
        ]


class SLHADumper(AbsDumper):
    """A dumper class for SLHA output."""

    def _update_line_option(self) -> None:
        self.line_option.block_str = self.config("block_str")
        self.line_option.decay_str = self.config("decay_str")
        self.line_option.comment = self.config("comments_preserve").keep_tail
        self.line_option.pre_comment = self.config("comments_preserve").keep_line

    def _read_config(self, sw: yaslha.config.SectionWrapper) -> None:
        super()._read_config(sw)
        self._config["separate_blocks"] = sw.getboolean("separate_blocks")
        self._config["forbid_last_linebreak"] = sw.getboolean("forbid_last_linebreak")
        self._config["document_blocks"] = sw.get_list("document_blocks")
        self._config["block_str"] = sw["block_str"]
        self._config["decay_str"] = sw["decay_str"]
        self._config["float_lower"] = sw.getboolean("float_lower")
        self._config["write_version"] = sw.getboolean("write_version")
        self._update_line_option()

    def config(self, k: str) -> Any:
        """Get a current value of configuration."""
        return super().config(k)

    def set_config(self, k: str, v: Any) -> None:
        """Set configuration."""
        # check before set
        pass
        # set
        super().set_config(k, v)
        # operations after set
        if k in ["block_str", "decay_str", "comments_preserve"]:
            self._update_line_option()

    def __init__(self, **kw: Any) -> None:
        self.line_option = yaslha.line.LineOutputOption()
        self._read_config(yaslha.cfg["SLHADumper"])
        for k, v in kw.items():
            self.set_config(k, v)

    def _version_comment(self) -> str:
        return "# written by {} {}".format(yaslha.__pkgname__, yaslha.__version__)

    def _version_comment_regexp(self) -> str:
        return r"^\s*#\s*written\s+by\s+{}\s+".format(yaslha.__pkgname__)

    def dump(self, slha: "yaslha.slha.SLHA") -> str:
        """Return SLHA-format text of an SLHA object."""
        document_blocks = [
            v.upper() for v in self.config("document_blocks")  # normalize to upper
        ]  # type: Sequence[str]

        lines = []  # type: List[str]
        for block in self._blocks_sorted(slha):
            lines.extend(
                self.dump_block(block, document_block=(block.name in document_blocks))
            )
            if self.config("separate_blocks"):
                lines.append("#")
        for decay in self._decays_sorted(slha):
            lines.extend(
                self.dump_block(decay, document_block=(decay.pid in document_blocks))
            )
            if self.config("separate_blocks"):
                lines.append("#")
        if self.config("separate_blocks"):
            lines.pop()
        if self.config("comments_preserve").keep_line:
            for c in slha.tail_comment:
                lines.append(format_comment(c, add_sharp=True, strip=False))

        # replace version string
        if self.config("write_version"):
            re_version = re.compile(self._version_comment_regexp())
            lines = [v for v in lines if not re_version.match(v)]
            lines.insert(0, self._version_comment())

        result = "\n".join(lines) + "\n"

        if self.config("forbid_last_linebreak"):
            result = result.rstrip()

        return result

    def dump_block(self, block, document_block=False):
        # type: (BlockLike, bool)->List[str]
        """Return SLHA-format text of a block."""
        lines = block.head.to_slha(self.line_option)
        for line in self._block_lines_ordered(block):
            lines.extend(line.to_slha(self.line_option))

        # special spacing for MODSEL block
        # because SDECAY somehow use (1x,i5,1x,i5,3x,a100).
        if isinstance(block, Block) and block.name == "MODSEL":
            re_modsel = re.compile(r"^\s*(\d+)\s*(\d+)\s*(#.*)?$")

            def reformat(match):
                # type: (re.Match[str])->str
                if match.group(3):
                    return " {:>5} {:>5}   {}".format(*match.groups())
                else:
                    return " {:>5} {:>5}   #".format(*match.groups())

            lines = [re_modsel.sub(reformat, i) for i in lines]

        # special spacing for MASS block
        if isinstance(block, Block) and block.name == "MASS":
            re_mass = re.compile(r"^\s*(\d+)")
            lines = [
                re_mass.sub(lambda x: " {:>9}".format(x.group(1)), i) for i in lines
            ]

        return self._document_out(lines) if document_block else lines


class AbsMarshalDumper(AbsDumper):
    """An abstract class for dumpers handling marshaled data."""

    SCHEME_VERSION = 3  # type: ClassVar[int]

    def _read_config(self, sw: yaslha.config.SectionWrapper) -> None:
        return super()._read_config(sw)

    def config(self, k: str) -> Any:
        """Get a current value of configuration."""
        return super().config(k)

    def set_config(self, k: str, v: Any) -> None:
        """Set configuration."""
        super().set_config(k, v)

    def _format_specification(self) -> Any:
        return OrderedDict(
            type="SLHA",
            formatter="{} {}".format(yaslha.__pkgname__, yaslha.__version__),
            scheme=self.SCHEME_VERSION,
        )

    def marshal(self, slha):
        # type: (yaslha.slha.SLHA)->Mapping[str, Any]
        """Return Mashaled object of an SLHA object."""
        blocks = OrderedDict()  # type: MutableMapping[str, Any]
        for block in self._blocks_sorted(slha):
            blocks[block.name] = self.marshal_block(block)
        decays = OrderedDict()  # type: MutableMapping[int, Any]
        for decay in self._decays_sorted(slha):
            decays[decay.pid] = self.marshal_block(decay)
        tail_comments = [format_comment(c, strip=False) for c in slha.tail_comment]

        result = OrderedDict()  # type: MutableMapping[str, Any]
        result["format"] = self._format_specification()
        if blocks:
            result["block"] = blocks
        if decays:
            result["decay"] = decays
        if self.config("comments_preserve").keep_line and tail_comments:
            result["tail_comments"] = tail_comments
        return result

    def marshal_block(self, block: "BlockLike") -> Mapping[str, Any]:
        """Return marshaled object of a block."""
        info = block.head.dump()
        value = []
        comment = []
        comment.extend(block.head._dump_comment())
        for line in self._block_lines_ordered(block):
            value.append(line.dump())
            comment.extend(line._dump_comment())

        comment = [
            c
            for c in comment
            if (c[0] == "pre" and self.config("comments_preserve").keep_line)
            or (c[0] != "pre" and self.config("comments_preserve").keep_tail)
        ]

        result = OrderedDict()  # type: MutableMapping[str, Any]
        if info:
            result["info"] = info
        if value:
            result["values"] = value
        if comment:
            result["comments"] = comment
        return result


class YAMLDumper(AbsMarshalDumper):
    """A dumper for YAML output."""

    def __init__(self, **kw: Any) -> None:
        self._read_config(yaslha.cfg["YAMLDumper"])
        for k, v in kw.items():
            self.set_config(k, v)

        self.yaml = ruamel.yaml.YAML()
        self.yaml.default_flow_style = None

        # we need not it is marked as omap (OrderedDict);
        # it could be just a dict as an output.
        # (but we may change the mind....)
        self.yaml.representer.yaml_representers[
            OrderedDict
        ] = self.yaml.representer.yaml_representers[dict]
        # # another idea...
        # def represent_list(self, data):
        #     flow_style = all(isinstance(i, str) or not hasattr(i, '__iter__')
        #                      for i in data)
        #     return self.represent_sequence(u'tag:yaml.org,2002:seq', data,
        #                                    flow_style=flow_style)
        # self.yaml.representer.yaml_representers[list] = represent_list

    def dump(self, slha: "yaslha.slha.SLHA") -> str:
        """Return YAML-format text of an SLHA object."""
        stream = ruamel.yaml.compat.StringIO()
        self.yaml.dump(self.marshal(slha), stream)
        return str(stream.getvalue())


class JSONDumper(AbsMarshalDumper):
    """A dumper for JSON output."""

    def __init__(self, **kw: Any) -> None:
        self._read_config(yaslha.cfg["JSONDumper"])
        for k, v in kw.items():
            self.set_config(k, v)

        self.indent = 2

    def dump(self, slha: "yaslha.slha.SLHA") -> str:
        """Return JSON-format text of an SLHA object."""
        return json.dumps(self.marshal(slha), indent=self.indent)



"""Block-like object of SLHA data."""


import logging
from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from typing import (
    Any,
    ClassVar,
    Generic,
    Iterator,
    List,
    MutableMapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

from yaslha._collections import OrderedTupleOrderInsensitiveDict
from yaslha.comment import CommentInterface
from yaslha.line import (
    BlockHeadLine,
    DecayHeadLine,
    DecayKeyType,
    DecayLine,
    DecayValueType,
    InfoKeyType,
    InfoLine,
    InfoValueType,
    KeyType,
    ValueLine,
    ValueType,
)

T = TypeVar("T")

logger = logging.getLogger(__name__)


KT = TypeVar("KT", KeyType, InfoKeyType)
VT = TypeVar("VT", ValueType, InfoValueType)
LT = TypeVar("LT", ValueLine, InfoLine)
CT = TypeVar("CT", str, List[str])
KTG = TypeVar("KTG", KeyType, InfoKeyType, DecayKeyType)


class GenericBlock(Generic[KTG, CT], metaclass=ABCMeta):
    """Block-like object containing comments."""

    @abstractmethod
    def __init__(self) -> None:
        self.head = NotImplemented  # type: Union[BlockHeadLine, DecayHeadLine]
        self._comment = CommentInterface(self)  # type: CommentInterface[KTG, CT]

    @property
    def comment(self) -> "CommentInterface[KTG, CT]":
        """Give the interface to comments."""
        return self._comment

    @abstractmethod
    def _get_comment(self, key: KTG) -> CT:
        pass

    @abstractmethod
    def _get_pre_comment(self, key: KTG) -> List[str]:
        pass

    @abstractmethod
    def _set_comment(self, key: KTG, value: CT) -> None:
        pass

    @abstractmethod
    def _set_pre_comment(self, key: KTG, value: Sequence[str]) -> None:
        pass

    @classmethod
    @overload
    def new(cls, obj: Union[int, DecayHeadLine]) -> "Decay":
        """..."""
        pass

    @classmethod  # noqa: F811
    @overload
    def new(cls, obj: Union[str, BlockHeadLine]) -> Union["Block", "InfoBlock"]:
        """..."""
        pass

    @classmethod  # noqa: F811
    def new(
        cls, obj: Union[int, str, DecayHeadLine, BlockHeadLine]
    ) -> "Union[Block, InfoBlock, Decay]":
        """Create a GenericBlock object according to the argument."""
        if isinstance(obj, int) or isinstance(obj, DecayHeadLine):
            return Decay(obj)
        name = obj.name if isinstance(obj, BlockHeadLine) else obj
        if name.endswith("INFO"):
            return InfoBlock(obj)
        else:
            return Block(obj)


class AbsBlock(GenericBlock[KT, CT], Generic[KT, VT, LT, CT], metaclass=ABCMeta):
    """Abstract class for SLHA blocks."""

    @abstractmethod
    def __init__(self, obj: Union[BlockHeadLine, str]) -> None:
        super().__init__()
        if isinstance(obj, BlockHeadLine):
            self.head = obj  # type: BlockHeadLine
        elif isinstance(obj, str):
            self.head = BlockHeadLine(name=obj)
        else:
            raise TypeError(obj)
        self._comment = CommentInterface(self)
        # _data must be initialized in subclasses
        self._data = NotImplemented  # type: Any

    @property
    def name(self) -> str:
        """Return the name of block (always in upper case)."""
        return self.head.name

    @property
    def q(self) -> Optional[float]:
        """Return the Q value."""
        return self.head.q

    @q.setter
    def q(self, value: Optional[float]) -> None:
        self.head.q = value

    @abstractmethod
    def __getitem__(self, key: KT) -> Union[VT, Sequence[VT]]:
        """Return the value corresponding to the key."""

    @abstractmethod
    def __setitem__(self, key: KT, value: VT) -> None:
        """Set the value for the key."""

    @abstractmethod
    def __delitem__(self, key: KT) -> None:
        """Delete the value for the key."""

    @abstractmethod
    def update_line(self, line: LT) -> None:
        """Add the line to the block, overriding if exists."""

    @abstractmethod
    def keys(self, sort: bool = False) -> Iterator[KT]:
        """Return the keys."""

    __iter__ = keys

    @abstractmethod
    def items(self, sort: bool = False) -> Iterator[Tuple[KT, VT]]:
        """Return (key, value) tuples."""

    @abstractmethod
    def _lines(self, sort: bool = False) -> Iterator[Tuple[KT, LT]]:
        pass


class Block(AbsBlock[KeyType, ValueType, ValueLine, str]):
    """SLHA block that has one value for one key."""

    def __init__(self, obj: Union[BlockHeadLine, str]) -> None:
        super().__init__(obj)
        self._data = OrderedDict()  # type: OrderedDict[KeyType, ValueLine]

    def __getitem__(self, key: KeyType) -> ValueType:
        """Return the value corresponding to the key."""
        return self._data[key].value

    def __setitem__(self, key: KeyType, value: ValueType) -> None:
        """Set the value for the key."""
        if key in self._data:
            self._data[key].value = value
        else:
            self._data[key] = ValueLine.new(key, value)

    def __delitem__(self, key: KeyType) -> None:
        """Delete the value for the key."""
        del self._data[key]

    def update_line(self, line: ValueLine) -> None:
        """Add the line to the block, overriding if exists."""
        self._data[line.key] = line

    def merge(self, another: "Union[Block, InfoBlock]") -> None:
        """Merge another block."""
        if isinstance(another, Block):
            for _, line in another._lines():
                self.update_line(line)
        else:
            raise ValueError(another)

    def get(self, *key: Any, default: T) -> Union[ValueType, T]:
        """Return the value for the key if exists, or default value."""
        # Here, `key` is always a tuple.
        # Meanwhile, the key for `__getitem__` is a tuple only if length > 1.
        if isinstance(key, tuple) and len(key) == 1:
            key = key[0]
        if key in self._data:
            return self.__getitem__(key)
        else:
            return default

    def keys(self, sort: bool = False) -> Iterator[KeyType]:
        """Return the keys."""
        for k, _ in self._lines(sort=sort):
            yield k

    __iter__ = keys

    def items(self, sort: bool = False) -> Iterator[Tuple[KeyType, ValueType]]:
        """Return (key, value) tuples."""
        for k, line in self._lines(sort=sort):
            yield k, line.value

    def _lines(self, sort: bool = False) -> Iterator[Tuple[KeyType, ValueLine]]:
        if sort:
            key_line_tuples = list(self._data.items())
            key_line_tuples.sort(key=lambda k: self._lines_sort_key(k[0]))
            for i in key_line_tuples:
                yield i
        else:
            for i in self._data.items():
                yield i

    @staticmethod
    def _lines_sort_key(k: KeyType) -> List[int]:
        if isinstance(k, int):
            return [k]
        elif isinstance(k, list):
            return k
        else:
            return []

    def _get_comment(self, key: KeyType) -> str:
        return self._data[key].comment

    def _get_pre_comment(self, key: KeyType) -> List[str]:
        return self._data[key].pre_comment

    def _set_comment(self, key: KeyType, value: Optional[str]) -> None:
        self._data[key].comment = value or ""

    def _set_pre_comment(self, key: KeyType, value: Optional[Sequence[str]]) -> None:
        self._data[key].pre_comment = [v for v in value] if value else []


class InfoBlock(AbsBlock[InfoKeyType, InfoValueType, InfoLine, List[str]]):
    """SLHA block that may have multiple values for one key."""

    def __init__(self, obj: Union[BlockHeadLine, str]) -> None:
        super().__init__(obj)
        self._data = []  # type: List[InfoLine]

    def __getitem__(self, key: InfoKeyType) -> Sequence[InfoValueType]:
        """Return the value corresponding to the key."""
        return tuple(line.value for line in self._data if line.key == key)

    def __setitem__(self, key: InfoKeyType, value: Sequence[InfoValueType]) -> None:
        """Set the value for the key."""
        if isinstance(value, str):
            raise TypeError(value)  # Fail-safe; only List[str] is allowed!
        self.__delitem__(key)
        self._data.extend([InfoLine(key, v) for v in value])

    def __delitem__(self, key: InfoKeyType) -> None:
        """Delete the value for the key."""
        self._data = [line for line in self._data if line.key != key]

    def update_line(self, line: InfoLine) -> None:
        """Add the line to the block, overriding if exists."""
        self.__delitem__(line.key)
        self._data.append(line)

    def append_line(self, line: InfoLine) -> None:
        """Add the line, appending to the existing one if exists."""
        self._data.append(line)

    def append(self, key: InfoKeyType, value: InfoValueType) -> None:
        """Append the value for the key."""
        self.append_line(InfoLine(key, value))

    def merge(self, another: "Union[Block, InfoBlock]") -> None:
        """Merge another block."""
        if isinstance(another, InfoBlock):
            updated = {}  # type: MutableMapping[InfoKeyType, bool]
            for key, line in another._lines():
                if updated.get(key):
                    self.append_line(line)
                else:
                    updated[key] = True
                    self.update_line(line)
        else:
            raise ValueError(another)

    def keys(self, sort: bool = False) -> Iterator[InfoKeyType]:
        """Return the keys."""
        keys_dict = {}  # type: MutableMapping[InfoKeyType, bool]
        for line in self._data:
            keys_dict[line.key] = True
        keys = list(keys_dict.keys())
        if sort:
            keys.sort()
        for k in keys:
            yield k

    __iter__ = keys

    def items(self, sort: bool = False) -> Iterator[Tuple[InfoKeyType, InfoValueType]]:
        """Return (key, value) tuples."""
        for key, line in self._lines(sort):
            yield key, line.value

    def _lines(self, sort: bool = False) -> Iterator[Tuple[InfoKeyType, InfoLine]]:
        for key in self.keys(sort):
            for line in self._data:
                if line.key == key:
                    yield key, line

    def _get_comment(self, key: InfoKeyType) -> List[str]:
        return [line.comment for line in self._data if line.key == key]

    def _get_pre_comment(self, key: InfoKeyType) -> List[str]:
        return self._data[key].pre_comment

    def _set_comment(self, key: InfoKeyType, value: Optional[Sequence[str]]) -> None:
        lines = [line for line in self._data if line.key == key]
        if value is None:
            value = []
        if len(lines) < len(value):
            raise ValueError(value)  # too many values
        for i, line in enumerate(lines):
            if i < len(value):
                line.comment = value[i]

    def _set_pre_comment(
        self, key: InfoKeyType, value: Optional[Sequence[str]]
    ) -> None:
        for line in self._data:
            if line.key == key:
                line.pre_comment = [v for v in value] if value else []
                value = []  # to remove all the remaining pre_comment


class Decay(GenericBlock[DecayKeyType, str]):
    """Decay block."""

    br_normalize_threshold = 1.0e-6  # type: ClassVar[float]

    def __init__(self, obj: Union[DecayHeadLine, int]) -> None:
        super().__init__()
        if isinstance(obj, DecayHeadLine):
            self.head = obj  # type: DecayHeadLine
        elif isinstance(obj, int):
            self.head = DecayHeadLine(pid=obj, width=0)
        else:
            raise TypeError(obj)
        self._data = (
            OrderedTupleOrderInsensitiveDict()
        )  # type: OrderedTupleOrderInsensitiveDict[DecayKeyType, DecayLine]

    @property
    def pid(self) -> int:
        """Return the pid of mother particle."""
        return self.head.pid

    @property
    def width(self) -> float:
        """Return the total width."""
        return self.head.width

    def update_line(self, line: DecayLine) -> None:
        """Add the line to the block, overriding if exists."""
        self._data[line.key] = line

    def br(self, *key: int) -> DecayValueType:
        """Return the BR of given channel."""
        if key in self._data:
            return self._data[key].br
        else:
            return 0

    def partial_width(self, *key: int) -> float:
        """Return the width of given channel."""
        return self.width * self.br(*key)

    def keys(self, sort: bool = False) -> Iterator[DecayKeyType]:
        """Return the keys."""
        for k, _ in self._lines(sort):
            yield k

    __iter__ = keys

    def items_br(self, sort=False):
        # type: (bool)->Iterator[Tuple[DecayKeyType, DecayValueType]]
        """Return (key, BR) tuples, sorted by the BR."""
        for k, line in self._lines(sort):
            yield k, line.br

    def items_partial_width(self, sort=False):
        # type: (bool)->Iterator[Tuple[DecayKeyType, float]]
        """Return (key, width) tuples, sorted by the BR."""
        for k, line in self._lines(sort):
            yield k, self.width * line.br

    def _lines(self, sort: bool = False) -> Iterator[Tuple[DecayKeyType, DecayLine]]:
        if sort:
            key_line_tuples = list(self._data.items())
            key_line_tuples.sort(key=lambda k: -k[1].br)
            for i in key_line_tuples:
                yield i
        else:
            for i in self._data.items():
                yield i

    def _get_comment(self, key: DecayKeyType) -> str:
        return self._data[key].comment

    def _get_pre_comment(self, key: DecayKeyType) -> List[str]:
        return self._data[key].pre_comment

    def _set_comment(self, key: DecayKeyType, value: Optional[str]) -> None:
        self._data[key].comment = value or ""

    def _set_pre_comment(
        self, key: DecayKeyType, value: Optional[Sequence[str]]
    ) -> None:
        self._data[key].pre_comment = [v for v in value] if value else []

    def normalize(self, force: bool = False) -> None:
        """Normalize the branching ratios.

        This method normalize all the branching ratios so that the sum becomes
        unity or less. In particular, if `force` is set True, they are
        normalized so that the sum becomes unity, regardless of the current
        value.

        If `force` is False and the sum is less than one, the branching ratio
        is not normalized, assuming that some decay channels are not listed. If
        `force` is False and the sum slightly exceeds the unity, the branching
        ratios are normalized, while if the excess is larger than
        `br_normalize_threshold`, `ValueError` is raised.
        """
        br_list = [br for _, br in self.items_br(sort=True)]
        total = sum(reversed(br_list))  # sum taken from smaller to reduce error

        if not total > 0:
            return  # stable particle
        elif force:
            logger.debug("BR for %d force normalized: %g", self.pid, total)
            pass  # to normalize
        elif total > 1 + self.br_normalize_threshold:
            logger.critical("BR for %d exceeds unity: %g", self.pid, total)
            raise ValueError(total)
        elif total >= 1:
            pass  # to normalize
        else:  # if force==False and total < 1
            rest = 1 - total
            assert rest > 0
            if rest > self.br_normalize_threshold:
                if hasattr(self, "_br_warned"):
                    pass
                else:
                    # warn only once
                    self._br_warned = True
                    logger.warning("BR for %d is less than unity by %g", self.pid, rest)
            return  # not normalize

        for v in self._data.values():
            v.br /= total

    def set_partial_width(self, *args: Union[int, float]) -> None:
        """Update the partial width and recalculate BRs of all channels."""
        if len(args) < 3 or not all(isinstance(i, int) for i in args[:-1]):
            raise KeyError(*args)
        key = cast(List[int], args[:-1])
        new_partial_width = float(args[-1])

        self.normalize()

        if key in self._data:
            old_partial_width = self.partial_width(*key)
        else:
            old_partial_width = 0
            self.update_line(DecayLine(br=0, channel=key))

        # update total width
        old_width = self.width
        new_width = new_partial_width - old_partial_width + old_width
        self.head.width = new_width

        # update the modified channel
        self._data[key].br = new_partial_width / new_width
        target = self._data[key]

        for line in self._data.values():
            if line != target:
                line.br *= old_width / new_width

    def remove(self, *key: int) -> None:
        """Remove the channel and recalculate BRs of all the other channels."""
        self.set_partial_width(*key, 0)
        del self._data[key]



"""Module of SLHA object class."""
import copy
import logging
from collections import OrderedDict
from typing import Any, List, Tuple, TypeVar, Union

from yaslha._collections import OrderedCaseInsensitiveDict as CIDict
from yaslha.block import AbsBlock, Block, Decay, InfoBlock
from yaslha.line import BlockHeadLine, DecayValueType, InfoLine, ValueLine, ValueType

BlockValueLine = Union[ValueLine, InfoLine]
SLHAItemValueType = Union[Block, Decay, ValueType, DecayValueType]
T = TypeVar("T")
logger = logging.getLogger(__name__)


class SLHA:
    """SLHA object, representing a SLHA-format text."""

    def __init__(self) -> None:
        self.blocks = CIDict()  # type: CIDict[str, Union[Block, InfoBlock]]
        self.decays = OrderedDict()  # type: OrderedDict[int, Decay]
        self.tail_comment = []  # type: List[str]

    def add_block(self, obj: Union["Block", "InfoBlock", "Decay"]) -> None:
        """Add a block to SLHA file.

        The name is automatically detected from the object.
        """
        if isinstance(obj, AbsBlock):
            self.blocks[obj.name] = obj
        elif isinstance(obj, Decay):
            self.decays[obj.pid] = obj
        else:
            raise TypeError

    @staticmethod
    def _key_reduce(key: Any) -> Tuple[Union[str, int], Any]:
        if isinstance(key, str) or isinstance(key, int):
            return key, None
        elif len(key) == 1:
            return key[0], None
        elif isinstance(key[0], str):
            return key[0], (key[1] if len(key) == 2 else key[1:])
        else:
            # decay block does not allow deep indexing
            raise KeyError(key)

    def __getitem__(self, key: Any) -> Any:
        """Get values of SLHA object or deeper.

        ``SLHA[str]`` and ``SLHA[int]`` give the specified block and decay
        block, respectively. For ordinary blocks, further referencing is
        possible as ``SLHA[str, *key]``, while decay blocks refuse such
        referencing for safety.
        """
        if isinstance(key, str):
            return self.blocks[key]
        elif isinstance(key, int):
            return self.decays[key]
        elif hasattr(key, "__len__"):
            if len(key) == 1:
                return self.__getitem__(key[0])
            elif len(key) >= 2 and isinstance(key[0], str):
                block = self.blocks[key[0]]
                return block[key[1] if len(key) == 2 else tuple(key[1:])]
        raise KeyError(key)

    def get(self, *key: Any, default: Any = None) -> Any:
        """Return the value if exists, or default."""
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def __setitem__(self, key: Any, value: Any) -> None:
        """Set a value of SLHA object or deeper."""
        if isinstance(key, str):
            assert isinstance(value, Block) or isinstance(value, InfoBlock)
            value.head.name = key  # correct the name of Block
            self.blocks[key] = value
        elif isinstance(key, int):
            assert isinstance(value, Decay)
            value.head.pid = key  # correct the pid of Decay
            self.decays[key] = value
        elif hasattr(key, "__len__") and len(key) >= 2 and isinstance(key[0], str):
            if key[0] not in self.blocks:
                self.add_block(AbsBlock.new(BlockHeadLine(name=key[0])))
            block = self.blocks[key[0]]
            block[key[1] if len(key) == 2 else tuple(key[1:])] = value
        else:
            raise KeyError(key)

    def __delitem__(self, key: Any) -> None:
        """Delete values of SLHA object or deeper."""
        if isinstance(key, str):
            del self.blocks[key]
        elif isinstance(key, int):
            del self.decays[key]
        elif hasattr(key, "__len__") and len(key) >= 2 and isinstance(key[0], str):
            block = self.blocks[key[0]]
            del block[key[1] if len(key) == 2 else tuple(key[1:])]
        else:
            raise KeyError(key)

    def normalize(self, blocks: bool = True, decays: bool = True) -> None:
        """Normalize the head-lines so that names/pids match the dict keys."""
        if blocks:
            for name, b in self.blocks.items():
                b.head.name = name
        if decays:
            for pid, d in self.decays.items():
                d.head.pid = pid

    def merge(self, another: "SLHA") -> None:
        """Merge another SLHA data into this object."""
        for name, block in another.blocks.items():
            self_block = self.blocks.get(name)
            if self_block:
                self_block.merge(block)
            else:
                self.blocks[name] = copy.deepcopy(block)
        self.decays.update(copy.deepcopy(another.decays))
        if another.tail_comment:
            self.tail_comment = copy.deepcopy(another.tail_comment)



"""Test for copy.deepcopy applied to SLHA object."""

import copy
import logging
import unittest

from pytest import approx

from yaslha.parser import SLHAParser

logger = logging.getLogger("test_info")


class TestCopy(unittest.TestCase):
    """Copy operation of SLHA data."""

    slha_string = """
Block SPINFO         # Program information
     1   SOFTSUSY    # spectrum calculator
     2   1.8.4       # version number
Block MODSEL  # Select model
     1    1   # sugra
Block alpha   # Effective Higgs mixing parameter
          -1.13716828e-01   # alpha
Block gauge Q= 4.64649125e+02
     1     3.60872342e-01   # g'(Q)MSSM drbar
     2     6.46479280e-01   # g(Q)MSSM drbar
     3     1.09623002e+00   # g3(Q)MSSM drbar
Block au Q= 4.64649125e+02
  3  3    -5.04995511e+02   # At(Q)MSSM drbar

DECAY   1000023     1.95831641E-02   # chi_20
     5.00000000E-01    2    -2000011        11
     4.00000000E-01    2    -2000013        13
     1.00000000E-01    3     1000022         1        -1

DECAY   999     1.00E-02    # data for testing
     0.10    2    123      4                 # 0.0010
     0.15    2    123      5                 # 0.0015
     0.30    2    123   -123                 # 0.0030
     0.40    3    123    123    123          # 0.0040
     0.05    4      1      2      3      4   # 0.0005
"""

    def setUp(self):
        parser = SLHAParser()
        self.slha = parser.parse(self.slha_string)

    def test_copy(self):
        assert self.slha["au", 3, 3] == -5.04995511e02
        assert self.slha["spinfo", 2] == ("1.8.4",)
        assert self.slha[999].partial_width(123, 123, 123) == approx(0.40 * 0.01)

        c = copy.copy(self.slha)

        assert c["au", 3, 3] == -5.04995511e02
        assert c["spinfo", 2] == ("1.8.4",)
        assert c[999].br(123, 123, 123) == 0.40

        c["au", 3, 3] = 1
        c["spinfo"].append(2, "another line")
        c[999].set_partial_width(123, 123, 123, 0.0)

        assert c["au", 3, 3] == 1
        assert c["spinfo", 2] == ("1.8.4", "another line")
        assert c[999].partial_width(123, 123, 123) == 0
        assert c[999].width == 0.006

        assert self.slha["au", 3, 3] == 1
        assert self.slha["spinfo", 2] == ("1.8.4", "another line")
        assert self.slha[999].partial_width(123, 123, 123) == 0
        assert self.slha[999].width == 0.006

    def test_deepcopy(self):
        assert self.slha["au", 3, 3] == -5.04995511e02
        assert self.slha["spinfo", 2] == ("1.8.4",)
        assert self.slha[999].partial_width(123, 123, 123) == 0.40 * 0.01

        c = copy.deepcopy(self.slha)

        assert c["au", 3, 3] == -5.04995511e02
        assert c["spinfo", 2] == ("1.8.4",)
        assert c[999].br(123, 123, 123) == 0.40

        c["au", 3, 3] = 1
        c["spinfo"].append(2, "another line")
        c[999].set_partial_width(123, 123, 123, 0.0)

        assert c["au", 3, 3] == 1
        assert c["spinfo", 2] == ("1.8.4", "another line")
        assert c[999].partial_width(123, 123, 123) == 0
        assert c[999].width == 0.006

        assert self.slha["au", 3, 3] == -5.04995511e02
        assert self.slha["spinfo", 2] == ("1.8.4",)
        assert self.slha[999].partial_width(123, 123, 123) == 0.40 * 0.01


# cspell:ignore softsusy modsel sminputs msbar drbar mgut mssm higgs hmix sugra tanb



"""Helpers for `yaslha.line` module."""

import re
from typing import Any, List, Sequence, TypeVar, Union, cast

import numpy

T = TypeVar("T", str, List[str])

# -----------------------------------------------------------------------------
# type definitions
# -----------------------------------------------------------------------------
"""Type for keys of ordinary blocks."""
KeyType = Union[None, int, Sequence[int]]
"""Type for values of ordinary blocks."""
ValueType = float  # typing.int is included in typing.float
"""Type for keys of info blocks."""
InfoKeyType = int
""""Type for values of INFO blocks."""
InfoValueType = str
"""Type for keys of decay blocks."""
DecayKeyType = Sequence[int]
"""Type for values of decay blocks."""
DecayValueType = float

# -----------------------------------------------------------------------------
# regular expressions
# -----------------------------------------------------------------------------
"""Regexp for float data in SLHA file."""
FLOAT = r"[+-]?(?:\d+\.\d*|\d+|\.\d+)(?:[deDE][+-]\d+)?"
"""Regexp for integer data in SLHA file."""
INT = r"[+-]?\d+"
"""Regexp for block names in SLHA file."""
NAME = r"[A-Za-z][A-Za-z0-9]*"
"""Regexp for values in INFO blocks."""
INFO = r"[^#]+"
"""Regexp for separators in SLHA file."""
SEP = r"\s+"
"""Regexp for tails of SLHA data line, capturing comment."""
TAIL = r"\s*(?:\# ?(?P<comment>.*))?"

"""Compiled regexp for float data in SLHA file."""
RE_INT = re.compile("^" + INT + "$")
"""Compiled regexp for int data in SLHA file."""
RE_FLOAT = re.compile("^" + FLOAT + "$")


def cap(regexp: str, name: str) -> str:
    """Return capture-pattern of REGEXP string."""
    return "(?P<{}>{})".format(name, regexp)


def possible(regexp: str) -> str:
    """Return possible pattern ```?``` of REGEXP string."""
    return "(?:{})?".format(regexp)


# -----------------------------------------------------------------------------
# other utility
# -----------------------------------------------------------------------------
def _float(obj: Any) -> float:
    """Convert any values to float if possible, otherwise raise an error."""
    if isinstance(obj, str):
        obj = obj.replace("d", "e").replace("D", "E")
    return float(obj)


def to_number(v: Any) -> float:
    """Convert any object to float or int depending on the expression."""
    if isinstance(v, float) or isinstance(v, int):
        return v
    elif isinstance(v, str):
        if RE_INT.match(v):
            return int(v)
        elif RE_FLOAT.match(v):
            return _float(v)
        raise ValueError("to_number failed: %s" % v)
    elif isinstance(v, numpy.ndarray) and v.ndim == 0:
        return cast(float, v.__pos__())
    else:
        return to_number(str(v))


def number_to_str(v: float, int_format: str = "d", float_format: str = "16.8e") -> str:
    """Convert int or float to string."""
    if isinstance(v, int):
        return ("{:" + int_format + "}").format(v)
    elif isinstance(v, float):
        return ("{:" + float_format + "}").format(v)
    else:
        raise TypeError(v)


def format_comment(comment: T, add_sharp: bool = True, strip: bool = True) -> T:
    """Format a comment string."""
    if isinstance(comment, str):
        comment = comment.strip() if strip else comment.rstrip()
        if add_sharp and not comment.startswith("#"):
            return "# " + comment if comment else "#"
        else:
            return comment
    else:
        return [format_comment(c, add_sharp, strip) for c in comment]



"""Sample codes of yaslha, serves also as test codes.

This example contains simple read/write of SLHA data.
"""

import logging
import unittest

import pytest

from yaslha.parser import SLHAParser

logger = logging.getLogger("test_info")


class TestExampleAdvanced(unittest.TestCase):
    """Simple read/write of SLHA data."""

    slha_string = """
# SUSY Les Houches Accord 1.0 - example spectrum file
# Info from spectrum calculator
Block SPINFO         # Program information
     1   SOFTSUSY    # spectrum calculator
     2   1.8.4       # version number
# Input parameters
Block MODSEL  # Select model
     1    1   # sugra
Block MINPAR  # SUSY breaking input parameters
     3    1.00000000e+01   # tanb(MZ) MSSM drbar
     2    2.50000000e+02   # m12(MGUT) MSSM drbar
     4    1.00000000e+00   # sign(mu(MGUT)) MSSM drbar
     1    1.00000000e+02   # m0(MGUT) MSSM drbar
     5   -1.00000000e+02   # A0(MGUT) MSSM drbar
#
# mgut=2.551299875e+16 GeV
Block MASS  # Mass spectrum
# PDG code     mass               particle
   1000001     5.73103437e+02   # ~d_L
   1000002     5.67658152e+02   # ~u_L
   1000003     5.73029886e+02   # ~s_L
   1000004     5.67583798e+02   # ~c_L
   1000005     5.15617364e+02   # ~b_1
   1000006     3.96457239e+02   # ~t_1
# Higgs mixing
Block alpha   # Effective Higgs mixing parameter
          -1.13716828e-01   # alpha
Block stopmix  # stop mixing matrix
  1  1     5.37975095e-01   # O_{11}
  1  2     8.42960733e-01   # O_{12}
  2  1     8.42960733e-01   # O_{21}
  2  2    -5.37975095e-01   # O_{22}
Block gauge Q= 4.64649125e+02
     1     3.60872342e-01   # g'(Q)MSSM drbar
     2     6.46479280e-01   # g(Q)MSSM drbar
     3     1.09623002e+00   # g3(Q)MSSM drbar
Block hmix Q= 4.64649125e+02  # Higgs mixing parameters
     1     3.58660361e+02   # mu(Q)MSSM drbar
     2     9.75139550e+00   # tan beta(Q)MSSM drbar
     3     2.44923506e+02   # higgs vev(Q)MSSM drbar
     4     1.69697051e+04   # [m3^2/cosB sinB](Q)MSSM drbar
Block au Q= 4.64649125e+02
  3  3    -5.04995511e+02   # At(Q)MSSM drbar
Block ad Q= 4.64649125e+02
  3  3    -7.97992485e+02   # Ab(Q)MSSM drbar
"""

    def setUp(self):
        parser = SLHAParser()
        self.slha = parser.parse(self.slha_string)

    def test_get(self):
        # default accessor raises KeyError if missing
        with pytest.raises(KeyError):
            self.slha["Ad"][1, 1]
        with pytest.raises(KeyError):
            self.slha["Ad", 1, 1]

        # get method allows to use default values
        assert self.slha["Ad"].get(3, 3, default=0) == -7.97992485e02
        assert self.slha["Ad"].get(2, 2, default=-123.456) == -123.456
        assert self.slha["Ad"].get(1, 1, default=None) is None

        assert self.slha.get("Au", 3, 3, default=0) == -5.04995511e02
        assert self.slha.get("Au", 2, 2, default=-1.0) == -1.0
        assert self.slha.get("Ad", 1, 1, default="NOTFOUND") == "NOTFOUND"

    def test_iterator_within_a_block(self):
        # block works as an iterator
        for key in self.slha["hmix"]:
            assert key in [1, 2, 3, 4]

        minpar_keys = list(self.slha["minpar"])
        assert minpar_keys == [3, 2, 4, 1, 5]  # order is saved

        alpha_keys = list(self.slha["alpha"])
        assert alpha_keys == [None]  # no-param line has None as a spurious key

        stopmix_keys = list(self.slha["stopmix"])
        assert stopmix_keys == [(1, 1), (1, 2), (2, 1), (2, 2)]  # multi-param as tuples

        # one can explicitly call Block.keys() method.
        assert list(self.slha["hmix"]) == [1, 2, 3, 4]
        assert list(self.slha["hmix"].keys()) == [1, 2, 3, 4]

        for k, v in self.slha["hmix"].items():
            assert v == self.slha["hmix", k]

        for k, v in self.slha["stopmix"].items():
            assert v == self.slha["stopmix"][k]
            assert (
                v == self.slha["stopmix", k]
            )  # you may notice this syntax is too tricky.

    def test_block_names(self):
        # SLHA.block_names() gives a generator for block names.
        expected_blocks = [
            "spinfo",
            "modsel",
            "minpar",
            "mass",
            "alpha",
            "stopmix",
            "gauge",
            "hmix",
            "au",
            "ad",
        ]
        for block_name in self.slha.blocks.keys():
            assert block_name.lower() in expected_blocks
            expected_blocks.remove(block_name.lower())
        assert len(expected_blocks) == 0


# cspell:ignore softsusy modsel sminputs msbar drbar mgut mssm higgs hmix sugra tanb



"""Module of a class to handle comment interface."""

import logging
from typing import TYPE_CHECKING, Generic, List, TypeVar, Union

from typing_extensions import Literal

from yaslha._line import format_comment

if TYPE_CHECKING:
    from yaslha.block import GenericBlock
    from yaslha._line import DecayKeyType, KeyType, InfoKeyType  # noqa: F401

KTG = TypeVar("KTG", "KeyType", "InfoKeyType", "DecayKeyType")
CT = TypeVar("CT", str, List[str])
HEAD = Literal["head"]

logger = logging.getLogger(__name__)


class CommentInterface(Generic[KTG, CT]):
    """Accessor object to the comments in blocks."""

    def __init__(self, block: "GenericBlock[KTG, CT]") -> None:
        self._block = block  # type: GenericBlock[KTG, CT]
        self._pre = PreCommentInterface(block)  # type: PreCommentInterface[KTG, CT]

    @property
    def pre(self) -> "PreCommentInterface[KTG, CT]":
        """Return pre-comment interface."""
        return self._pre

    def __getitem__(self, key: Union[KTG, HEAD]) -> Union[CT, str]:  # noqa: F811
        """Return comment."""
        if key == "head":
            return self._block.head.comment
        else:
            return self._block._get_comment(key)

    def __setitem__(self, key: Union[KTG, HEAD], value: Union[CT, str, None]) -> None:
        """Set comment."""
        if key == "head":
            self._block.head.comment = value or ""
        else:
            self._block._set_comment(key, value)

    def __call__(self, key: Union[KTG, HEAD], **kw: bool) -> Union[CT, str]:
        """Return comment in specified format."""
        return format_comment(self.__getitem__(key), **kw)


class PreCommentInterface(Generic[KTG, CT]):
    """Accessor object to the pre-line comments in blocks."""

    def __init__(self, block: "GenericBlock[KTG, CT]"):
        self._block = block  # type: GenericBlock[KTG, CT]

    def __getitem__(self, key: Union[KTG, HEAD]) -> List[str]:
        """Return pre-line comment."""
        if key == "head":
            return self._block.head.pre_comment
        else:
            return self._block._get_pre_comment(key)

    def __setitem__(self, key: Union[KTG, HEAD], value: Union[List[str], None]) -> None:
        """Set pre-line comment."""
        if key == "head":
            self._block.head.pre_comment = value or []
        else:
            self._block._set_pre_comment(key, value)

    def __call__(self, key: Union[KTG, HEAD], **kw: bool) -> List[str]:
        """Return pre-line comment in specified format."""
        return format_comment(self.__getitem__(key), **kw)



import collections
from abc import ABCMeta, abstractmethod
from typing import Any, Optional, TypeVar, Union, overload

K = TypeVar("K")
V = TypeVar("V")
T = TypeVar("T", None, Any)

class _OrderedNormalizedDict(collections.OrderedDict[K, V], metaclass=ABCMeta):
    @classmethod
    @abstractmethod
    def _n(self, key: K) -> K: ...
    def __init__(self, *args: Any, **kwds: Any) -> None: ...
    def __setitem__(self, key: K, value: V) -> None: ...
    def __getitem__(self, key: K) -> V: ...
    def __delitem__(self, key: K) -> None: ...
    def __contains__(self, key: Any) -> bool: ...
    @overload
    def get(self, key: K) -> Optional[V]: ...
    @overload
    def get(self, key: Any, default: Union[V, T]) -> Union[V, T]: ...
    @overload
    def pop(self, key: K) -> V: ...
    @overload
    def pop(self, key: K, default: Union[V, T] = ...) -> Union[V, T]: ...
    def move_to_end(self, key: K, last: bool = True) -> None: ...

class OrderedCaseInsensitiveDict(_OrderedNormalizedDict[K, V]):
    @classmethod
    def _n(self, key: K) -> K: ...

class OrderedTupleOrderInsensitiveDict(_OrderedNormalizedDict[K, V]):
    @classmethod
    def _n(self, key: K) -> K: ...



"""Sample codes of yaslha, serves also as test codes.

This example contains block-wise operations.
"""

import copy
import logging
import unittest

import pytest

from yaslha.block import Block, Decay
from yaslha.parser import SLHAParser

logger = logging.getLogger("test_info")


class TestExampleBlockOperation(unittest.TestCase):
    """Block-wise operation for SLHA data."""

    slha_string = """
Block SPINFO         # Program information
     1   SOFTSUSY    # spectrum calculator
     2   1.8.4       # version number
Block MODSEL  # Select model
     1    1   # sugra
Block alpha   # Effective Higgs mixing parameter
          -1.13716828e-01   # alpha
Block gauge Q= 4.64649125e+02
     1     3.60872342e-01   # g'(Q)MSSM drbar
     2     6.46479280e-01   # g(Q)MSSM drbar
     3     1.09623002e+00   # g3(Q)MSSM drbar
Block au Q= 4.64649125e+02
  3  3    -5.04995511e+02   # At(Q)MSSM drbar

DECAY   1000023     1.95831641E-02   # chi_20
     5.00000000E-01    2    -2000011        11
     4.00000000E-01    2    -2000013        13
     1.00000000E-01    3     1000022         1        -1

DECAY   999     1.00E-02    # data for testing
     0.10    2    123      4                 # 0.0010
     0.15    2    123      5                 # 0.0015
     0.30    2    123   -123                 # 0.0030
     0.40    3    123    123    123          # 0.0040
     0.05    4      1      2      3      4   # 0.0005
"""

    def setUp(self):
        parser = SLHAParser()
        self.slha = parser.parse(self.slha_string)

    def test_basic_block_operation_1(self):
        gauge = self.slha["gauge"]

        # now one can handle this block itself
        gauge[1] = 0.1
        gauge[2] = 0.4
        gauge.q = 100.123
        del gauge[3]

        # the original SLHA data **is** affected.
        assert self.slha["gauge"][1] == 0.1
        assert self.slha["gauge", 2] == 0.4
        assert self.slha["gauge"].q == 100.123
        with pytest.raises(KeyError):
            self.slha["gauge", 3]

    def test_basic_block_operation_2(self):
        # if you want to unlink the block from SLHA, you must use deep-copy.
        gauge_copied = copy.deepcopy(self.slha["Gauge"])

        assert gauge_copied[1] == 3.60872342e-01
        assert gauge_copied.q == 4.64649125e02  # the data and q value are copied

        # changing the contents of the new block
        gauge_copied[2] = -0.2
        del gauge_copied[3]

        # the new block is modified
        assert gauge_copied[2] == -0.2
        with pytest.raises(KeyError):
            gauge_copied[3]

        # the original SLHA data is not modified.
        assert self.slha["gauge", 1] == 3.60872342e-01
        assert self.slha["gauge", 2] == 6.46479280e-01
        assert self.slha["gauge", 3] == 1.09623002e00

        # you may replace the block
        self.slha["gauge"] = gauge_copied
        assert self.slha["gauge", 1] == 3.60872342e-01
        assert self.slha["gauge", 2] == -0.2
        assert self.slha.get("gauge", 3) is None

    def test_slha_iterators_blocks(self):
        # iterator only for ordinal blocks
        expected = ["spinfo", "modsel", "alpha", "gauge", "au"]
        for name in self.slha.blocks:
            assert name.lower() in expected
            expected.remove(name.lower())
        assert len(expected) == 0

    def test_slha_iterators_decays(self):
        # iterator only for decay blocks
        expected = [1000023, 999]
        for pid in self.slha.decays:
            assert pid in expected
            expected.remove(pid)
        assert len(expected) == 0

    def test_add_block(self):
        # a new block
        new_block = Block("")
        new_block.q = 123.456
        new_block[3] = 10.05

        # assign a new block to SLHA object
        self.slha["NEWBLOCK"] = new_block

        assert self.slha["newblock", 3] == 10.05
        assert self.slha["newblock"].q == 123.456

        # the block name is modified accordingly
        assert self.slha["NewBlock"].name.lower() == "newblock"

    def test_add_block_2(self):
        # another method to add block
        new_block = Block("newblock")
        new_block.q = 123.456
        new_block[3] = 10.05

        self.slha.add_block(new_block)
        assert self.slha["newblock", 3] == 10.05
        assert self.slha["newblock"].q == 123.456

    def test_add_decay_block(self):
        # a new block
        new_decay = Decay(789)
        new_decay.set_partial_width(123, 123, 0.01)
        new_decay.set_partial_width(123, -123, 0.02)
        new_decay.set_partial_width(4, 3, 2, 0.01)

        # assign a new block to SLHA object
        self.slha[789] = new_decay

        assert self.slha[789].br(123, 123) == 0.25
        assert self.slha[789].partial_width(123, -123) == 0.02
        assert self.slha[789].width == 0.04

    def test_delete_block(self):
        # remove a block
        del self.slha["modsel"]

        with pytest.raises(KeyError):
            self.slha["modsel"]

        # remove a decay block
        del self.slha[999]
        assert list(self.slha.decays.keys()) == [1000023]


# cspell:ignore softsusy modsel sminputs msbar drbar mgut mssm higgs hmix sugra tanb



"""Sample codes of yaslha, serves also as test codes.

This example contains basic read/write of SLHA data.
"""

import logging
import unittest

import pytest

from yaslha.parser import SLHAParser

logger = logging.getLogger("test_info")


class TestExampleBasic(unittest.TestCase):
    """Basic read/write of SLHA data."""

    slha_string = """
# SUSY Les Houches Accord 1.0 - example input file
# Snowmass point 1a
Block MODSEL  # Select model
     1    1   # sugra
Block SMINPUTS   # Standard Model inputs
     3      0.1172  # alpha_s(MZ) SM msbar
     5      4.25    # Mb(mb) SM msbar
     6    174.3     # Mtop(pole)
Block MINPAR  # SUSY breaking input parameters
     3     10.0     # tanb
     4      1.0     # sign(mu)
     1    100.0     # m0
     2    250.0     # m12
     5   -100.0     # A0
"""

    def setUp(self):
        parser = SLHAParser()
        self.slha = parser.parse(self.slha_string)

    def test_read_1(self):
        # simply read the data from Block
        assert self.slha["MODSEL", 1] == 1
        assert self.slha["SMINPUTS", 3] == 0.1172
        assert self.slha["MINPAR", 3] == 10.0

        # block name is case-insensitive
        assert self.slha["SMinputs", 6] == 174.3

    def test_read_2(self):
        # read the data in another syntax
        assert self.slha["MODSEL"][1] == 1
        assert self.slha["SMinputs"][6] == 174.3

        # these are equivalent to, for example,
        minpar_block = self.slha["minpar"]
        assert minpar_block[3] == 10.0
        assert minpar_block[4] == 1.0

    def test_update(self):
        # update/add data
        self.slha["SMinputs", 6] = 172.3
        self.slha["SMINPUTS", 11] = 0.511e-3

        assert self.slha["SMINPUTS", 6] == 172.3
        assert self.slha["SMINPUTS"][11] == 0.511e-3

        # a bit more technical operation
        minpar_block = self.slha["minpar"]
        minpar_block[3] = 40.0
        minpar_block[6] = 123.456

        assert self.slha["minpar", 3] == 40.0
        assert self.slha["minpar", 6] == 123.456

    def test_add_block(self):
        # To create a new block
        self.slha["extpar", 1] = 100.12
        assert self.slha["EXTPAR", 1] == 100.12

        # This doesn't work because self.slha["new"] is not yet defined (thus KeyError).
        with pytest.raises(KeyError):
            self.slha["new"][1] = 100.12

    def test_delete(self):
        # you can remove the data
        del self.slha["SMinputs", 6]
        del self.slha["minpar"][3]

        with pytest.raises(KeyError):
            self.slha["minpar", 3]

        with pytest.raises(KeyError):
            self.slha["sminputs"][6]

        # block structure is not removed even if all the data are removed
        del self.slha["minpar", 1]
        self.slha["minpar"][1] = 2
        assert self.slha["minpar", 1] == 2


# cspell:ignore modsel sugra tanb sminputs msbar mtop



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yaslha', 'yaslha.tests', 'yaslha.tests.examples']

package_data = \
{'': ['*'],
 'yaslha.tests': ['data/*', 'data/pylha_json/*', 'data/pylha_yaml/*']}

install_requires = \
['click>=8.0,<9.0',
 'colorama>=0.4.4,<0.5.0',
 'coloredlogs>=15.0,<16.0',
 'numpy>=1.20,<2.0',
 'ruamel.yaml>=0.17.4,<0.18.0',
 'typing-extensions>=4.0,<5.0']

entry_points = \
{'console_scripts': ['yaslha = yaslha.script:main']}

setup_kwargs = {
    'name': 'yaslha',
    'version': '0.3.2',
    'description': 'A Python package to handle SLHA (SUSY Les Houches Accord) files.',
    'long_description': '[![Build Status](https://github.com/misho104/yaslha/actions/workflows/unit-test.yaml/badge.svg?branch=master)](https://github.com/misho104/yaslha/actions/workflows/unit-test.yaml)\n[![Coverage Status](https://codecov.io/gh/misho104/yaslha/branch/master/graph/badge.svg?token=SL19A4WATT)](https://codecov.io/gh/misho104/yaslha)\n[![Typecheck Status](https://github.com/misho104/yaslha/actions/workflows/type-check.yaml/badge.svg?branch=master)](https://github.com/misho104/yaslha/actions/workflows/type-check.yaml)\n[![Doc Status](http://readthedocs.org/projects/yaslha/badge/)](https://yaslha.readthedocs.io/)\n[![PyPI version](https://badge.fury.io/py/yaslha.svg)](https://badge.fury.io/py/yaslha)\n[![License: MIT](https://img.shields.io/badge/License-MIT-ff25d1.svg)](https://github.com/misho104/yaslha/blob/master/LICENSE)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n\n[yaslha](https://github.com/misho104/yaslha): Yet Another SLHA module for Python3\n=================================================================================\n\nA Python3 package to manipulate [SLHA](http://skands.physics.monash.edu/slha/) files and convert them to other file formats (JSON and YAML).\n\nQuick Start\n-----------\n\n(TBW)\n\nIntroduction\n------------\n\nThe [SUSY Les Houches Accord](http://skands.physics.monash.edu/slha/) is a data format widely used in particle physics phenomenology.\nIt is originally defined in [arXiv:hep-ph/0311123](https://arxiv.org/abs/hep-ph/0311123) and extended to SLHA2 in [arXiv:0801.0045](https://arxiv.org/abs/0801.0045).\n\nBecause of its birth in FORTRAN era it is a fixed length format such as\n\n```\nBLOCK SMINPUTS                  # Standard Model input parameters\n     1     1.27934000e+02   # alpha_em^-1(M_Z)^MSbar\n     2     1.16637000e-05   # G_F [GeV^-2]\n     3     1.17200000e-01   # alpha_S(M_Z)^MSbar\n     4     9.11876000e+01   # mZ (pole)\n     5     4.18000000e+00   # mb(mb)^MSbar\n     6     1.73300000e+02   # mt (pole)\n     7     1.77682000e+00   # mtau (pole)\n#\nBLOCK ALPHA                    #\n          -2.68630018e-02   # Higgs mixing parameter\n#\nBLOCK HMIX Q= 2.00000000e+02   # Higgs parameters (DRbar)\n     1     5.40000000e+02   # mu(Q)\n     2     4.00000000e+01   # tanbeta(Q)\n     3     2.46220569e+02   # vev(Q)\n     4     2.30400000e+05   # mA^2(Q)\n#\n...\n```\n\nand extended in many program codes.\n\nPython has two famous SLHA parser: [PySLHA](http://www.insectnation.org/projects/pyslha) by Andy Buckley and [pylha](https://github.com/DavidMStraub/pylha) by David M. Straub.\n[yaslha](https://github.com/misho104/yaslha) is "yet another" SLHA parser, influenced much by these two parsers.\n\nPython regrettably experienced a terrible era due to the transition from Python2 to Python3.\nTo reduce code complexity, this package supports only Python3.4 and later versions.\n\nUsage\n-----\n\n(TBW)\n\nAuthor\n------\n\n[Sho Iwamoto / Misho](https://www.misho-web.com/), under much influence from [PySLHA](http://www.insectnation.org/projects/pyslha) by Andy Buckley and [pylha](https://github.com/DavidMStraub/pylha) by David M. Straub.',
    'author': 'Sho Iwamoto (Misho)',
    'author_email': 'webmaster@misho-web.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/misho104/yaslha',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0',
}


setup(**setup_kwargs)



"""Utility module."""

from collections import OrderedDict, defaultdict
from typing import List, MutableMapping, Sequence, TypeVar, Union

from yaslha._line import KeyType

T = TypeVar("T", int, KeyType)


BLOCKS_DEFAULT_ORDER = [
    "SPINFO",
    "DCINFO",
    "MODSEL",
    "SMINPUTS",
    "MINPAR",
    "EXTPAR",
    "VCKMIN",
    "UPMNSIN",
    "MSQ2IN",
    "MSU2IN",
    "MSD2IN",
    "MSL2IN",
    "MSE2IN",
    "TUIN",
    "TDIN",
    "TEIN",
    "MASS",
    "NMIX",
    "UMIX",
    "VMIX",
    "ALPHA",
    "FRALPHA",
    "HMIX",
    "GAUGE",
    "MSOFT",
    "MSQ2",
    "MSU2",
    "MSD2",
    "MSL2",
    "MSE2",
    "STOPMIX",
    "SBOTMIX",
    "STAUMIX",
    "USQMIX",
    "DSQMIX",
    "SELMIX",
    "SNUMIX",
    "AU",
    "AD",
    "AE",
    "TU",
    "TD",
    "TE",
    "YU",
    "YD",
    "YE",
]  # type: Sequence[str]


def sort_blocks_default(block_names: Sequence[str]) -> List[str]:
    """Sort block names according to specified order."""
    result = []
    block_names = [n.upper() for n in block_names]
    peeked = OrderedDict([(k, False) for k in block_names])

    for name in [n.upper() for n in BLOCKS_DEFAULT_ORDER]:
        if name in block_names:
            result.append(name)
            peeked[name] = True

    return result + [k for k, v in peeked.items() if not v]


def sort_pids_default(pids: Sequence[T]) -> List[Union[T, int]]:
    """Sort block names according to specified order."""
    tmp = defaultdict(list)  # type: MutableMapping[str, List[int]]
    non_int = []  # type: List[Union[T, int]]

    for i in pids:
        if not isinstance(i, int):
            non_int.append(i)  # fail safe
            continue

        j = i % 1000000
        if i < 1000000:
            tmp["01_sm"].append(i)
        elif i >= 3000000:
            tmp["10_others"].append(i)
        elif i == 1000021:
            tmp["02_gluino"].append(i)
        elif j <= 6:
            if j % 2:
                tmp["04_sq-down"].append(i)
            else:
                tmp["03_sq-up"].append(i)
        elif i in [1000022, 1000023, 1000025, 1000035]:
            tmp["05_neut"].append(i)
        elif i in [1000024, 1000037]:
            tmp["06_char"].append(i)
        elif j in [11, 13, 15]:
            tmp["07_slep"].append(i)
        elif j in [12, 14, 16]:
            tmp["08_snu"].append(i)
        else:
            tmp["09_susy"].append(i)
    # tricks to avoid mypy errors
    pids = list(pid for group in sorted(tmp.keys()) for pid in sorted(tmp[group]))
    non_int = pids + non_int
    return non_int



"""Sample codes of yaslha, serves also as test codes.

This example contains comment handlings.
"""

import logging
import unittest

import pytest

from yaslha.parser import SLHAParser

logger = logging.getLogger("test_info")


class TestExampleComment(unittest.TestCase):
    """Advanced read/write of SLHA data."""

    slha_string = """
# SLHA 1.0
# calculator
Block SPINFO         # Program information
     1   SOFTSUSY    # name
     2   1.8.4       # version number
# Input parameters
Block MODSEL  # Select model
     1    1   # sugra
Block MINPAR  # SUSY breaking input parameters
# comment between lines (1)
     1    1.00000000e+02   # m0(MGUT) MSSM drbar
     2    2.50000000e+02   # m12(MGUT) MSSM drbar
# comment between lines (2)
     3    1.00000000e+01   # tanb(MZ) MSSM drbar
     4    1.00000000e+00   # sign(mu(MGUT)) MSSM drbar
     5   -1.00000000e+02   # A0(MGUT) MSSM drbar
# comment after minpar
Block MASS  # Mass spectrum
# PDG code     mass               particle
   1000001     5.73103437e+02   # ~d_L
   1000002     5.67658152e+02   # ~u_L
   1000003     5.73029886e+02   # ~s_L
   1000004     5.67583798e+02   # ~c_L
   1000005     5.15617364e+02   # ~b_1
   1000006     3.96457239e+02   # ~t_1
Block hmix Q= 4.64649125e+02
     1     3.58660361e+02
     2     9.75139550e+00
     3     2.44923506e+02
     4     1.69697051e+04
#
#         PDG        Width
DECAY   1000023     1.95831641E-02   # chi_20
#    BR                NDA      ID1      ID2
     3.38444885E-02    2    -2000011        11   # BR(chi_20 -> ~e_R+ e- )
     3.38444885E-02    2     2000011       -11   # BR(chi_20 -> ~e_R- e+ )
     3.50457690E-02    2    -2000013        13   # BR(chi_20 -> ~mu_R+ mu- )
     3.50457690E-02    2     2000013       -13   # BR(chi_20 -> ~mu_R- mu+ )
     4.29284412E-01    2    -1000015        15   # BR(chi_20 -> ~tau_1+ tau- )
     4.29284412E-01    2     1000015       -15   # BR(chi_20 -> ~tau_1- tau+ )
     2.07507330E-06    2     1000022        22   # BR(chi_20 -> chi_10 photon )
#    BR                NDA      ID1      ID2       ID3
     1.75499602E-04    3     1000022         2        -2   # BR(chi_20 -> chi_10 u u )
     1.75229451E-04    3     1000022         4        -4   # BR(chi_20 -> chi_10 c c )
     2.33949042E-04    3     1000022         1        -1   # BR(chi_20 -> chi_10 d d )
#comment at SLHA-tail
"""

    def setUp(self):
        parser = SLHAParser()
        self.slha = parser.parse(self.slha_string)

    def test_read_1(self):
        # simple examples of pre-head, head, and line comments

        # pre-head comment is always List[str]
        assert self.slha["modsel"].comment.pre["head"] == ["# Input parameters"]
        assert self.slha["spinfo"].comment.pre["head"] == ["# SLHA 1.0", "# calculator"]

        # head comment is always str
        assert self.slha["modsel"].comment["head"] == "Select model"
        assert self.slha["spinfo"].comment["head"] == "Program information"

        # line-tail comment depends on block type
        assert self.slha["modsel"].comment[1] == "sugra"  # str for ordinary block
        assert self.slha["mass"].comment[1000001] == "~d_L"
        assert self.slha["spinfo"].comment[1] == ["name"]  # List[str] for INFO block
        assert self.slha["spinfo"].comment[2] == ["version number"]

        # another method to access
        modsel = self.slha["modsel"]
        assert modsel.comment.pre["head"] == ["# Input parameters"]
        assert modsel.comment["head"] == "Select model"
        assert modsel.comment[1] == "sugra"

        # comments between lines are recognized as "pre" comments of the next lines.
        minpar = self.slha["minpar"]
        assert minpar.comment.pre[1] == [
            "# comment between lines (1)"
        ]  # always List[str]
        assert minpar.comment.pre[3] == ["# comment between lines (2)"]

        # comments after block are recognized as pre-block comments of the next blocks.
        assert self.slha["mass"].comment.pre["head"] == ["# comment after minpar"]

        # the comment at the end of file is specially treated
        assert self.slha.tail_comment == ["#comment at SLHA-tail"]  # always List[str]

        # empty string (or empty list) is returned if comment does not exist.
        assert self.slha["hmix"].comment.pre["head"] == []
        assert self.slha["hmix"].comment["head"] == ""
        assert self.slha["hmix"].comment.pre[1] == []
        assert self.slha["hmix"].comment[1] == ""

    def test_read_2(self):
        # for decay blocks
        assert self.slha[1000023].comment.pre["head"] == [
            "#",
            "#         PDG        Width",
        ]
        assert self.slha[1000023].comment.pre[11, -2000011] == [
            "#    BR                NDA      ID1      ID2"
        ]
        assert self.slha[1000023].comment[11, -2000011] == "BR(chi_20 -> ~e_R+ e- )"

    def test_update(self):
        modsel = self.slha["modsel"]
        # line comments are a string.
        modsel.comment["head"] = "new head comment"
        modsel.comment[1] = "new comment 2"
        assert self.slha["modsel"].comment["head"] == "new head comment"
        assert self.slha["modsel"].comment[1] == "new comment 2"

        # pre-line comments are a list of string.
        modsel.comment.pre["head"] = ["new pre-head comment"]
        modsel.comment.pre[1] = ["new comment 1"]
        assert self.slha["modsel"].comment.pre["head"] == ["new pre-head comment"]
        assert self.slha["modsel"].comment.pre[1] == ["new comment 1"]

        # for INFO blocks, line comments are a list of string.
        assert self.slha["spinfo"].comment[1] == ["name"]
        assert self.slha["spinfo"].comment[2] == ["version number"]
        self.slha["spinfo"].comment[1] = ["new comment 1"]
        self.slha["spinfo"].comment[2] = ["new comment 2"]
        assert self.slha["spinfo"].comment[1] == ["new comment 1"]
        assert self.slha["spinfo"].comment[2] == ["new comment 2"]

        # add comments
        self.slha["minpar", 6] = 0.5
        self.slha["minpar", 7] = 0.1
        self.slha["minpar"].comment.pre[6] = ["extra parameters"]
        self.slha["minpar"].comment[6] = "extra 1"
        self.slha["minpar"].comment[7] = "extra 2"
        assert self.slha["minpar"].comment.pre[6] == ["extra parameters"]
        assert self.slha["minpar"].comment[6] == "extra 1"
        assert self.slha["minpar"].comment[7] == "extra 2"

        # one cannot assign comments to non-existing lines
        with pytest.raises(KeyError):
            self.slha["minpar"].comment[8] = "non-existing line comment"
        with pytest.raises(KeyError):
            self.slha["minpar"].comment.pre[8] = "non-existing pre-line comment"

    def test_delete(self):
        # comments can be removed by assigning None or empty values.
        self.slha["minpar"].comment[1] = None
        self.slha["minpar"].comment.pre[1] = None
        self.slha["minpar"].comment.pre["head"] = None
        self.slha["minpar"].comment["head"] = None

        assert self.slha["minpar"].comment[1] == ""
        assert self.slha["minpar"].comment.pre[1] == []
        assert self.slha["minpar"].comment.pre["head"] == []
        assert self.slha["minpar"].comment["head"] == ""  # note the different types


# cspell:ignore softsusy modsel sminputs msbar drbar mgut mssm higgs hmix sugra tanb



"""Sample codes of yaslha, serves also as test codes.

This example contains how to handle DECAY blocks.
"""

import logging
import unittest

from yaslha.parser import SLHAParser

logger = logging.getLogger("test_info")


class TestExampleDecayBlocks(unittest.TestCase):
    """Usage of DECAY blocks."""

    slha_string = """
Block SPINFO         # Program information
     1   SOFTSUSY    # spectrum calculator
     2   1.8.4       # version number
Block DCINFO          # Program information
     1    SDECAY       # Decay package
     2    1.0          # version number
Block gauge Q= 4.64649125e+02
     1     3.60872342e-01   # g'(Q)MSSM drbar
     2     6.46479280e-01   # g(Q)MSSM drbar
     3     1.09623002e+00   # g3(Q)MSSM drbar

#         PDG           Width
DECAY   1000023     1.95831641E-02   # chi_20
#    BR                NDA      ID1      ID2
     3.38444885E-02    2    -2000011        11   # BR(chi_20 -> ~e_R+ e- )
     3.38444885E-02    2     2000011       -11   # BR(chi_20 -> ~e_R- e+ )
     3.50457690E-02    2    -2000013        13   # BR(chi_20 -> ~mu_R+ mu- )
     3.50457690E-02    2     2000013       -13   # BR(chi_20 -> ~mu_R- mu+ )
     4.29284412E-01    2    -1000015        15   # BR(chi_20 -> ~tau_1+ tau- )
     4.29284412E-01    2     1000015       -15   # BR(chi_20 -> ~tau_1- tau+ )
     2.07507330E-06    2     1000022        22   # BR(chi_20 -> chi_10 photon )
#    BR                NDA      ID1      ID2       ID3
     1.75499602E-04    3     1000022         2        -2   # BR(chi_20 -> chi_10 u u )
     1.75229451E-04    3     1000022         4        -4   # BR(chi_20 -> chi_10 c c )
     2.33949042E-04    3     1000022         1        -1   # BR(chi_20 -> chi_10 d d )

DECAY   999     1.00E-02    # data for testing
     0.10    2    123      4                 # 0.0010
     0.15    2    123      5                 # 0.0015
     0.30    2    123   -123                 # 0.0030
     0.40    3    123    123    123          # 0.0040
     0.05    4      1      2      3      4   # 0.0005
"""

    def setUp(self):
        parser = SLHAParser()
        self.slha = parser.parse(self.slha_string)

    def test_read(self):
        # One can read decay blocks by the pid.
        n2_decay = self.slha[1000023]

        assert n2_decay.width == 1.95831641e-2  # total width
        assert n2_decay.br(-2000011, 11) == 3.38444885e-02
        assert n2_decay.br(11, -2000011) == 3.38444885e-02  # order is irrelevant

        expected_partial_width = 1.95831641e-2 * 3.38444885e-2
        assert n2_decay.partial_width(11, -2000011) == expected_partial_width

    def test_iterator_read(self):
        # decay block as an iterator
        for _key in self.slha[1000023]:
            pass

        for i, key in enumerate(self.slha[999].keys()):
            if i == 0:
                assert sorted(key) == [4, 123]
            elif i == 1:
                assert sorted(key) == [5, 123]
            elif i == 2:
                assert sorted(key) == [-123, 123]
            elif i == 3:
                assert sorted(key) == [123, 123, 123]
            else:
                assert sorted(key) == [1, 2, 3, 4]

        for i, key in enumerate(self.slha[999].keys(sort=True)):
            # the keys are sorted by the branding ratio descending.
            if i == 0:
                assert sorted(key) == [123, 123, 123]
            elif i == 1:
                assert sorted(key) == [-123, 123]
            elif i == 2:
                assert sorted(key) == [5, 123]
            elif i == 3:
                assert sorted(key) == [4, 123]
            else:
                assert sorted(key) == [1, 2, 3, 4]

        # branching ratio iterator
        for key, value in self.slha[999].items_br():
            assert value == self.slha[999].br(*key)

        # partial width iterator
        for key, value in self.slha[999].items_partial_width():
            assert value == self.slha[999].partial_width(*key)

    def test_update_and_remove(self):
        # modification is made through partial width, not by BRs.
        decay = self.slha[999]
        decay.set_partial_width(123, 4, 0.005)

        assert decay.partial_width(123, 4) == 0.005
        assert decay.width == 0.014
        assert decay.br(123, 4) == 0.005 / 0.014

        self.slha[999].set_partial_width(123, 123, 0.006)
        assert self.slha[999].width == 0.020
        assert self.slha[999].br(123, 123) == 0.006 / 0.020
        assert decay.br(123, 4) == 0.005 / 0.020

        self.slha[999].remove(123, 123)
        decay.remove(4, 3, 2, 1)
        decay.remove(5, 123)
        assert self.slha[999].width == 0.012
        assert decay.br(123, -123) == 0.003 / 0.012
        # cspell:ignore softsusy sdecay mssm drbar



"""Tests for convert sub-commmand."""

import logging
import pathlib
import re
import traceback
import unittest

import coloredlogs
from click.testing import CliRunner

import yaslha.dumper
from yaslha.script import convert

logger = logging.getLogger("test_info")


def check_and_separate_output(result):
    """Separate STDERR-like lines from STDOUT text."""
    if result.exit_code != 0:
        traceback.print_tb(result.exc_info[2])
        print(result.exc_info[1])
    assert result.exit_code == 0

    # separate logging lines to STDERR
    re_logline = re.compile(r"(yaslha\.\w+:)? (CRITICAL|ERROR|WARNING|DEBUG|INFO)[: ]")
    stdout = []
    stderr = []
    for i in result.output.splitlines():
        if re_logline.match(i):
            stderr.append(i)
        else:
            stdout.append(i)
    return stdout, stderr


def compare_lines(a, b):
    """Compare two texts line by line."""
    if isinstance(a, str):
        return compare_lines(a.splitlines(), b)
    if isinstance(b, str):
        return compare_lines(a, b.splitlines())
    a.append("*END_OF_TEXT*")
    b.append("*END_OF_TEXT*")
    na, nb = len(a), len(b)
    for i, ta in enumerate(a):
        assert ta == (b[i] if i < nb else "")
    assert na == nb


class TestConverter(unittest.TestCase):
    """Test class for converter sub-command."""

    def setUp(self):
        coloredlogs.set_level(40)
        self.data_dir = pathlib.Path(__file__).parent / "data"
        self.inputs = [
            str(path) for path in self.data_dir.glob("*.*") if path.is_file()
        ]
        self.runner = CliRunner()

    def test_idempotence(self):
        for input_file in self.inputs:
            for block_order in yaslha.dumper.BlocksOrder:
                for value_order in yaslha.dumper.ValuesOrder:
                    for comment in yaslha.dumper.CommentsPreserve:
                        args = [
                            "--input-type=SLHA",
                            "--output-type=SLHA",
                            "--blocks=" + block_order.name,
                            "--values=" + value_order.name,
                            "--comments=" + comment.name,
                        ]
                        result1 = self.runner.invoke(convert, args + [input_file])
                        result1_output, result1_stderr = check_and_separate_output(
                            result1
                        )
                        result2 = self.runner.invoke(
                            convert, args, input="\n".join(result1_output)
                        )
                        result2_output, result2_stderr = check_and_separate_output(
                            result2
                        )
                        compare_lines(result1_output, result2_output)



"""Module to describe each type of lines.

The line hierarchy is given by:

- AbsLine
  - BlockHeadLine
  - DecayHeadLine
  - InfoLine
  - ValueLine
    - NoIndexLine
    - OneIndexLine
    - TwoIndexLine
    - ThreeIndexLine
  - DecayLine
  - CommentLine
"""
import collections.abc as abc
import logging
import re
from abc import ABCMeta, abstractmethod
from typing import (
    Any,
    ClassVar,
    List,
    Optional,
    Pattern,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)

from yaslha._line import (
    FLOAT,
    INFO,
    INT,
    NAME,
    SEP,
    TAIL,
    DecayKeyType,
    DecayValueType,
    InfoKeyType,
    InfoValueType,
    KeyType,
    ValueType,
    _float,
    cap,
    format_comment,
    possible,
    to_number,
)

logger = logging.getLogger(__name__)

OS = Union[str, None]
SInt = Union[str, int]
SFloat = Union[str, float]
SValue = Union[str, ValueType]

CT = TypeVar("CT", str, List[str])
LT = TypeVar("LT", bound="AbsLine")
LT2 = TypeVar("LT2", bound="ValueLine")


class LineOutputOption:
    """Class to hold all the options on dumping lines."""

    # type comment for python 3.6
    # block_str: str  # the string "BLOCK" for block head
    # decay_str: str  # the string "DECAY" for decay-block head
    # comment: bool  # whether to output line-end comments
    # pre_comment: bool  # whether to output pre-line comments
    # float_lower: bool  # letter E for float numbers

    def __init__(self) -> None:
        self.block_str = "Block"
        self.decay_str = "Decay"
        self.comment = True
        self.pre_comment = True
        self.float_lower = False


class AbsLine(metaclass=ABCMeta):
    """Abstract class for SLHA-line like objects."""

    output_option = LineOutputOption()  # type: ClassVar[LineOutputOption]

    _pattern = NotImplemented  # type: ClassVar[str]
    _pattern_compiled = None  # type: ClassVar[Optional[Pattern[str]]]

    @classmethod
    def pattern(cls) -> Pattern[str]:
        """Return a regexp pattern matching a SLHA line for the type."""
        if cls._pattern_compiled is None:
            cls._pattern_compiled = re.compile("^{}$".format(cls._pattern), re.I)
        return cls._pattern_compiled

    @classmethod
    def construct(cls: Type[LT], line: str) -> "Optional[LT]":
        """Construct an object from a line if it maches the pattern."""
        match = cls.pattern().match(line)
        if match:
            return cls(**match.groupdict())
        else:
            return None

    @abstractmethod
    def __init__(self, **kwargs: Any) -> None:
        self.comment = NotImplemented  # type: str
        self.pre_comment = NotImplemented  # type: List[str]

    # from/to object/string representation
    def __str__(self) -> str:
        return self._to_slha(self.output_option)

    def to_slha(self, option: Optional[LineOutputOption] = None) -> List[str]:
        """Return the lines for SLHA-format, including pre-comments."""
        opt = option or self.output_option
        lines = self._format_pre_comment(opt)  # pre comment
        lines.append(self._to_slha(opt))  # line itself
        return lines

    def dump(self) -> List[SFloat]:
        """Return a list representing the line."""
        return self._dump()

    @classmethod
    def from_dump(cls: Type[LT], dump: Sequence[Any], **kw: Any) -> LT:
        """Return an object from dumped value."""
        return cls._from_dump(dump, **kw)

    # implementation
    @abstractmethod
    def _to_slha(self, opt: LineOutputOption) -> str:
        pass

    @abstractmethod
    def _dump(self) -> List[SFloat]:
        pass

    @classmethod
    @abstractmethod
    def _from_dump(cls: Type[LT], dump: Sequence[Any], **kw: Any) -> LT:
        pass

    # comment handling (type declaration for python 3.6)
    # comment: str
    # pre_comment: List[str]

    def _format_comment(self, opt: LineOutputOption) -> str:
        """Return the comment formatted for SLHA line."""
        if opt.comment:
            return format_comment(self.comment, add_sharp=True, strip=True)
        else:
            return "#"

    def _format_pre_comment(self, opt: LineOutputOption) -> List[str]:
        """Return the pre-comment formatted as SLHA lines."""
        if opt.pre_comment:
            return [
                format_comment(c, add_sharp=True, strip=False) for c in self.pre_comment
            ]
        else:
            return []

    @abstractmethod
    def _dump_comment(self) -> List[List[SFloat]]:
        pass

    # helper method
    @classmethod
    def _num_to_str(cls, opt, v, allow_int=False):
        # type: (LineOutputOption, float, bool)->str
        if isinstance(v, int) and allow_int:
            return str(v)
        f = "{:16.8e}" if opt.float_lower else "{:16.8E}"
        return f.format(v)


class BlockHeadLine(AbsLine):
    """Line for block header."""

    _pattern = (
        "Block"
        + SEP
        + cap(NAME, "name")
        + possible(SEP + r"Q=\s*" + cap(FLOAT, "q"))
        + TAIL
    )

    def __init__(self, name, q=None, comment=None):
        # type: (str, Optional[SFloat], OS)->None
        self.name = name
        self.q = None if q is None else _float(q)
        self.comment = comment or ""
        self.pre_comment = []

    @property
    def name(self) -> str:
        """Return name in upper case."""
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value.upper()

    def _to_slha(self, opt: LineOutputOption) -> str:
        if self.q is None:
            name = " {:17}   ".format(self.name)
        else:
            name = " {} Q={}   ".format(self.name, self._num_to_str(opt, self.q))
        return (opt.block_str + name + self._format_comment(opt)).rstrip()

    def _dump(self) -> List[SFloat]:
        return [] if self.q is None else ["Q=", self.q]

    @classmethod
    def _from_dump(cls: Type[LT], dump: Sequence[Any], **kw: Any) -> LT:
        name = str(kw.get("name"))
        if not name or len(kw) > 1:
            raise ValueError(kw)
        if len(dump) == 0:
            return cls(name=name)
        elif len(dump) == 2:
            if isinstance(dump[0], str) and dump[0].strip().upper() == "Q=":
                return cls(name=name, q=_float(dump[1]))
        raise ValueError(dump)

    def _dump_comment(self) -> List[List[SFloat]]:
        comments = [
            ["pre", "HEAD", c] for c in self.pre_comment
        ]  # type: List[List[SFloat]]
        if self.comment:
            comments.append(["HEAD", self.comment])
        return comments


class DecayHeadLine(AbsLine):
    """A line with format ``('DECAY',1x,I9,3x,1P,E16.8,0P,3x,'#',1x,A)``."""

    _pattern = "Decay" + SEP + cap(INT, "pid") + SEP + cap(FLOAT, "width") + TAIL

    def __init__(self, pid: SInt, width: SFloat, comment: OS = None) -> None:
        self.pid = int(pid)
        self.width = _float(width)
        self.comment = comment or ""
        self.pre_comment = []

    def _to_slha(self, opt: LineOutputOption) -> str:
        return "{decay} {pid:>9}   {width}   {comment}".format(
            decay=opt.decay_str,
            pid=self.pid,
            width=self._num_to_str(opt, self.width),
            comment=self._format_comment(opt),
        )

    def _dump(self) -> List[SFloat]:
        return [self.width]

    @classmethod
    def _from_dump(cls: Type[LT], dump: Sequence[Any], **kw: Any) -> LT:
        pid = kw.get("pid")
        if not pid or len(kw) > 1:
            raise ValueError(kw)
        elif len(dump) == 1:
            return cls(pid=int(pid), width=_float(dump[0]))
        raise ValueError(dump)

    def _dump_comment(self) -> List[List[SFloat]]:
        comments = [
            ["pre", "HEAD", c] for c in self.pre_comment
        ]  # type: List[List[SFloat]]
        if self.comment:
            comments.append(["HEAD", self.comment])
        return comments


class InfoLine(AbsLine):
    """Class for lines of INFO blocks.

    An info-block line is given by ``format(1x,I5,3x,A)``, which is not
    exclusive and matches other patterns. This class accept multi-lines, and
    thus values are List[str] and comments are multi-lined string, internally
    kept as List[str].
    """

    _pattern = r"\s*" + cap(INT, "key") + SEP + cap(INFO, "value") + TAIL

    def __init__(self, key, value, comment=None):
        # type: (InfoKeyType, InfoValueType, OS)->None
        self.key = int(key)  # type: InfoKeyType
        self.value = value.rstrip()  # type: InfoValueType
        self.comment = comment or ""
        self.pre_comment = []

    def _to_slha(self, opt: LineOutputOption) -> str:
        return " {:5d}   {:16}   {}".format(
            self.key, self.value, self._format_comment(opt)
        )

    def _dump(self) -> List[SFloat]:
        return [self.key, self.value]

    @classmethod
    def _from_dump(cls: Type["InfoLine"], dump: Sequence[Any], **kw: Any) -> LT:
        if kw:
            raise ValueError(kw)
        if len(dump) == 2:
            return cast(LT, cls(int(dump[0]), value=str(dump[1])))
        raise ValueError(dump)

    def _dump_comment(self) -> List[List[SFloat]]:
        comments = [
            ["pre", self.key, c] for c in self.pre_comment
        ]  # type: List[List[SFloat]]
        if self.comment:
            comments.append([self.key, self.comment])
        return comments


class ValueLine(AbsLine, metaclass=ABCMeta):
    """Abstract class for value lines in ordinary blocks."""

    @abstractmethod
    def __init__(self, key: KeyType, value: SValue, comment: OS = None) -> None:
        self.key = key  # type: KeyType
        self.value = to_number(value)  # type: ValueType
        self.comment = comment or ""
        self.pre_comment = []

    @classmethod
    def new(cls: Type[LT2], key: KeyType, value: SValue, comment: OS = None) -> LT2:
        """Construct line object according to the type of key.

        The resulting object is an instance of a subclass of `ValueLine` but
        not `DecayLine`.
        """
        if key is None:
            return cast(LT2, NoIndexLine(value, comment))
        elif isinstance(key, int):
            return cast(LT2, OneIndexLine(key, value, comment))
        elif isinstance(key, abc.Sequence) and all(isinstance(i, int) for i in key):
            if len(key) == 1:
                return cast(LT2, OneIndexLine(key[0], value, comment))
            elif len(key) == 2:
                return cast(LT2, TwoIndexLine(key[0], key[1], value, comment))
            elif len(key) == 3:
                return cast(LT2, ThreeIndexLine(key[0], key[1], key[2], value, comment))
        raise TypeError(key)

    def _to_slha(self, opt: LineOutputOption) -> str:
        if self.key is None:
            k = "      "
        elif isinstance(self.key, int):
            k = " {:5d}".format(self.key)
        else:
            k = "{:6}".format("".join(" {:2d}".format(k) for k in self.key))
        if isinstance(self.value, int):
            v = "{:10}      ".format(self.value)
        else:
            v = self._num_to_str(opt, self.value, False)
        return "{}   {}   {}".format(k, v, self._format_comment(opt))

    def _dump(self) -> List[SFloat]:
        if self.key is None:
            return [self.value]
        elif isinstance(self.key, int):
            return [self.key, self.value]
        elif isinstance(self.key, abc.Sequence):
            return [*self.key, self.value]
        raise TypeError(self.key)

    @classmethod
    def _from_dump(cls: Type[LT2], dump: Sequence[Any], **kw: Any) -> LT2:
        if kw:
            raise ValueError(kw)
        if len(dump) == 0:
            raise ValueError(dump)
        elif len(dump) == 1:
            return cls.new(key=None, value=dump[-1])
        elif len(dump) == 2:
            return cls.new(key=int(dump[0]), value=dump[-1])
        else:
            return cls.new(key=tuple(int(k) for k in dump[:-1]), value=dump[-1])

    def _dump_comment(self) -> List[List[SFloat]]:
        if self.key is None:
            key_tuple = []  # type: Sequence[int]
        elif isinstance(self.key, int):
            key_tuple = [self.key]
        elif isinstance(self.key, abc.Sequence):
            key_tuple = self.key
        comments = [
            ["pre", *key_tuple, c] for c in self.pre_comment
        ]  # type: List[List[SFloat]]
        if self.comment:
            comments.append([*key_tuple, self.comment])
        return comments


class NoIndexLine(ValueLine):
    """A line with ``format(9x, 1P, E16.8, 0P, 3x, '#', 1x, A)``."""

    _pattern = r"\s*" + cap(FLOAT, "value") + TAIL

    def __init__(self, value, comment=None):
        # type: (SValue, OS)->None
        super().__init__(None, value, comment)


class OneIndexLine(ValueLine):
    """A line with ``format(1x,I5,3x,1P,E16.8,0P,3x,'#',1x,A)``."""

    _pattern = r"\s*" + cap(INT, "i") + SEP + cap(FLOAT, "value") + TAIL

    def __init__(self, i, value, comment=None):
        # type: (SInt, SValue, OS)->None
        super().__init__(int(i), value, comment)


class TwoIndexLine(ValueLine):
    """A line with ``format(1x,I2,1x,I2,3x,1P,E16.8,0P,3x,'#',1x,A)``."""

    _pattern = (
        r"\s*"
        + cap(INT, "i1")
        + SEP
        + cap(INT, "i2")
        + SEP
        + cap(FLOAT, "value")
        + TAIL
    )

    def __init__(self, i1, i2, value, comment=None):
        # type: (SInt, SInt, SValue, OS)->None
        super().__init__((int(i1), int(i2)), value, comment)


class ThreeIndexLine(ValueLine):
    """A line with ``format(1x,I2,1x,I2,1x,I2,3x,1P,E16.8,0P,3x,'#',1x,A)``."""

    _pattern = (
        r"\s*"
        + cap(INT, "i1")
        + SEP
        + cap(INT, "i2")
        + SEP
        + cap(INT, "i3")
        + SEP
        + cap(FLOAT, "value")
        + TAIL
    )

    def __init__(self, i1, i2, i3, value, comment=None):
        # type: (SInt, SInt, SInt, SValue, OS)->None
        super().__init__((int(i1), int(i2), int(i3)), value, comment)


class DecayLine(ValueLine):
    """A decay line ``(3x,1P,E16.8,0P,3x,I2,3x,N (I9,1x),2x,'#',1x,A)``."""

    _pattern = (
        r"\s*"
        + cap(FLOAT, "br")
        + SEP
        + cap(INT, "nda")
        + SEP
        + cap(r"[0-9\s+-]+", "channel")
        + TAIL
    )

    def __init__(self, br, channel, nda=None, comment=None):
        # type: (Union[str, DecayValueType], Union[str, DecayKeyType], Any, OS)->None
        if isinstance(channel, str):
            self.key = tuple(int(p) for p in re.split(r"\s+", channel.strip()))
        else:
            self.key = channel  # type: DecayKeyType
        self.value = _float(br)  # type: DecayValueType

        self.comment = comment or ""
        self.pre_comment = []

    # provide synonym
    @property
    def br(self) -> DecayValueType:
        """Return the branching ratio."""
        return self.value

    @br.setter
    def br(self, br: DecayValueType) -> None:
        self.value = br

    def _to_slha(self, opt: LineOutputOption) -> str:
        pids = "".join("{:9d} ".format(pid) for pid in self.key)
        return "   {}   {:2d}   {}  {}".format(
            self._num_to_str(opt, self.br),
            len(self.key),
            pids,
            self._format_comment(opt),
        )

    def _dump(self) -> List[SFloat]:
        result = [self.br, len(self.key)]  # type: List[SFloat]
        result.extend(self.key)
        return result

    @classmethod
    def _from_dump(cls: Type["DecayLine"], dump: Sequence[Any], **kw: Any) -> LT2:
        if kw:
            raise ValueError(kw)
        if len(dump) >= 4:
            br, nda = _float(dump[0]), int(dump[1])
            pids = tuple(int(p) for p in dump[2:])
            if nda == len(pids):
                return cast(LT2, cls(br=br, channel=pids))
        raise ValueError(dump)

    def _dump_comment(self) -> List[List[SFloat]]:
        comments = [["pre", c] for c in self.pre_comment]  # type: List[List[SFloat]]
        if self.comment:
            comments.append([self.comment])
        for c in comments:
            c.extend(self.key)
        return comments


class CommentLine(AbsLine):
    """Comment line.

    This object is prepared for compatibility and not intended to be inserted
    in blocks or decay-blocks; therefore dumping methods are not implemented.
    """

    _pattern = r"\s*(?P<comment>\#.*)"

    def __init__(self, comment: OS = None) -> None:
        self.comment = (comment or "").rstrip()
        self.pre_comment = []

    def _to_slha(self, opt: LineOutputOption) -> str:
        raise NotImplementedError

    def _dump(self) -> List[SFloat]:
        raise NotImplementedError

    @classmethod
    def _from_dump(cls: Type[LT], dump: Sequence[Any], **kw: Any) -> LT:
        raise NotImplementedError

    def _dump_comment(self) -> List[List[SFloat]]:
        raise NotImplementedError



"""Sample codes of yaslha, serves also as test codes.

This example contains simple read/write of SLHA data.
"""

import logging
import unittest

import pytest

from yaslha.parser import SLHAParser

logger = logging.getLogger("test_info")


class TestExampleSimple(unittest.TestCase):
    """Simple read/write of SLHA data."""

    slha_string = """
# SUSY Les Houches Accord 1.0 - example spectrum file
# Info from spectrum calculator
Block SPINFO         # Program information
     1   SOFTSUSY    # spectrum calculator
     2   1.8.4       # version number
     3   Error Message 1
     3   Error Message 2
# Input parameters
Block MODSEL  # Select model
     1    1   # sugra
Block MINPAR  # SUSY breaking input parameters
     1    1.00000000e+02   # m0(MGUT) MSSM drbar
     2    2.50000000e+02   # m12(MGUT) MSSM drbar
     3    1.00000000e+01   # tanb(MZ) MSSM drbar
     4    1.00000000e+00   # sign(mu(MGUT)) MSSM drbar
     5   -1.00000000e+02   # A0(MGUT) MSSM drbar
#
# mgut=2.551299875e+16 GeV
Block MASS  # Mass spectrum
# PDG code     mass               particle
   1000001     5.73103437e+02   # ~d_L
   1000002     5.67658152e+02   # ~u_L
   1000003     5.73029886e+02   # ~s_L
   1000004     5.67583798e+02   # ~c_L
   1000005     5.15617364e+02   # ~b_1
   1000006     3.96457239e+02   # ~t_1
# Higgs mixing
Block alpha   # Effective Higgs mixing parameter
          -1.13716828e-01   # alpha
Block stopmix  # stop mixing matrix
  1  1     5.37975095e-01   # O_{11}
  1  2     8.42960733e-01   # O_{12}
  2  1     8.42960733e-01   # O_{21}
  2  2    -5.37975095e-01   # O_{22}
Block gauge Q= 4.64649125e+02
     1     3.60872342e-01   # g'(Q)MSSM drbar
     2     6.46479280e-01   # g(Q)MSSM drbar
     3     1.09623002e+00   # g3(Q)MSSM drbar
Block hmix Q= 4.64649125e+02  # Higgs mixing parameters
     1     3.58660361e+02   # mu(Q)MSSM drbar
     2     9.75139550e+00   # tan beta(Q)MSSM drbar
     3     2.44923506e+02   # higgs vev(Q)MSSM drbar
     4     1.69697051e+04   # [m3^2/cosB sinB](Q)MSSM drbar
Block au Q= 4.64649125e+02
  3  3    -5.04995511e+02   # At(Q)MSSM drbar
Block ad Q= 4.64649125e+02
  3  3    -7.97992485e+02   # Ab(Q)MSSM drbar
"""

    def setUp(self):
        parser = SLHAParser()
        self.slha = parser.parse(self.slha_string)

    def test_read_1(self):
        # Reading data of multi-param blocks
        assert self.slha["STOPMIX", 1, 1] == 5.37975095e-01
        assert self.slha["AU"][3, 3] == -5.04995511e02

        # To read data of no-param block, a dummy parameter None must be used.
        #  (usually "Block ALPHA" is only the case)
        assert self.slha["ALPHA", None] == -1.13716828e-01
        assert self.slha["ALPHA"][None] == -1.13716828e-01  # this syntax also works

        # reading the "Q" value of a block
        assert self.slha["GAUGE"].q == 4.64649125e02

    def test_update(self):
        self.slha["STOPMIX"][1, 1] = 0.5
        self.slha["AU", 1, 1] = 0.001
        self.slha["ALPHA", None] = 1.0

        assert self.slha["STOPMIX", 1, 1] == 0.5
        assert self.slha["AU"][1, 1] == 0.001
        assert self.slha["ALPHA"][None] == 1.0

        # create new block with multi params
        self.slha["lambda", 1, 2, 3] = 0.01
        assert self.slha["LAMBDA", 1, 2, 3] == 0.01

        # the "Q" value
        self.slha["GAUGE"].q = 400.0
        assert self.slha["GAUGE"].q == 400.0

        # Q value can be set to any blocks (even if physically invalid).
        self.slha["MASS"].q = 123.45
        assert self.slha["mass"].q == 123.45

    def test_delete(self):
        # multi-param or no-param
        del self.slha["STOPMIX", 1, 1]
        del self.slha["Alpha"][None]

        with pytest.raises(KeyError):
            self.slha["stopmix", 1, 1]

        with pytest.raises(KeyError):
            self.slha["Alpha", None]

        # removing Q-value
        self.slha["gauge"].q = None
        assert self.slha["gauge"].q is None

    def test_info_block(self):
        # blocks with names ending with "INFO" is specially treated.
        assert self.slha["spinfo"][1] == ("SOFTSUSY",)
        assert self.slha["spinfo"][2] == ("1.8.4",)
        assert self.slha["spinfo"][3] == ("Error Message 1", "Error Message 2")

        # all the contents must be a list of string.
        self.slha["dcinfo", 1] = ["DecayProgramName"]

        with pytest.raises(TypeError):
            self.slha["dcinfo", 2] = 1.0

        self.slha["dcinfo", 3] = ["Error message."]
        self.slha["dcinfo", 4] = ["Warning 1", "Warning 2"]

        assert self.slha["DCINFO"][3] == ("Error message.",)
        assert self.slha["DCINFO"][4] == ("Warning 1", "Warning 2")

        # append a value
        self.slha["dcinfo"].append(4, "Warning 3")

        assert len(self.slha["dcinfo"][4]) == 3


# cspell:ignore softsusy modsel sminputs msbar drbar mgut mssm higgs hmix sugra tanb



"""Definitions of customized collection classes."""

from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from typing import Any, Generic, TypeVar, Union, cast

K = TypeVar("K")
V = TypeVar("V")
T = TypeVar("T", None, Any)

_not_specified = object()


class _OrderedNormalizedDict(OrderedDict, Generic[K, V], metaclass=ABCMeta):
    """Abstract class for normalized OrderedDict.

    Normalization is given by the class method `_n`.
    """

    @classmethod
    @abstractmethod
    def _n(self, key: K) -> K:
        pass

    def __init__(self, *args: Any, **kwds: Any) -> None:
        """Initialize an ordered dictionary."""
        # first construct a temporal dict
        tmp = OrderedDict(*args, **kwds)
        for k, v in tmp.items():
            self.__setitem__(k, v)

    def __setitem__(self, key: K, value: V) -> None:
        super().__setitem__(self._n(key), value)

    def __getitem__(self, key: K) -> V:
        return cast(V, super().__getitem__(self._n(key)))

    def __delitem__(self, key: K) -> None:
        super().__delitem__(self._n(key))

    def __contains__(self, key: Any) -> bool:
        return super().__contains__(self._n(key))

    # def get(self, k: _KT, default: Union[_VT_co, _T]) -> Union[_VT_co, _T]: ...

    def get(self, key: Any, default: Union[V, T] = None) -> Union[V, T]:
        return self[key] if key in self else default

    def pop(self, key: K, default: Union[V, T, object] = _not_specified) -> Union[V, T]:
        if default == _not_specified:
            return cast("V", super().pop(self._n(key)))
        else:
            return super().pop(self._n(key), default=default)

    def move_to_end(self, key: K, last: bool = True) -> None:
        return super().move_to_end(self._n(key), last)


class OrderedCaseInsensitiveDict(_OrderedNormalizedDict[K, V]):
    """OrderedDict with case-insensitive keys.

    Keys are identified as case-insensitive. For a tuple as a key, the elements
    of the tuples are not normalized and remain case-sensitive.
    """

    @classmethod
    def _n(self, key: K) -> K:
        return key.upper() if hasattr(key, "upper") else key  # type: ignore


class OrderedTupleOrderInsensitiveDict(_OrderedNormalizedDict[K, V]):
    """OrderedDict with neglecting order of tuple elements.

    The identification is applied only if ```isinstance(key, tuple)``` is True,
    and only to the top-level elements.
    """

    @classmethod
    def _n(self, key: K) -> K:
        return tuple(sorted(key)) if isinstance(key, tuple) else key  # type: ignore
